/*****************************************************************************/
/*     Generator - Sega Genesis emulation - (c) James Ponder 1997-1998       */
/*****************************************************************************/
/*                                                                           */
/* compile.c                                                                 */
/*                                                                           */
/*****************************************************************************/

#define OLD_TYPEDEFS
typedef float float32;
typedef double float64;

#include "generator.h"
#include "registers.h"
#include "vdp.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <proto/exec.h> /* AmigaOS */

#define RECT_ALIGN (32) /* PowerPC */

void *rect_malloc(uint32 size, uint32 flags)
{
  uint8 *buf;

  buf = (uint8 *)AllocVec(size + (RECT_ALIGN - 1) + sizeof(void **), flags);
  if (buf != NULL) {
    uint8 *align_buf;

    memset(buf, 0, size + (RECT_ALIGN - 1) + sizeof(void **));

    align_buf = buf + (RECT_ALIGN - 1) + sizeof(void **);
    align_buf -= (uint32)align_buf & (RECT_ALIGN - 1);
    *(((void **)align_buf) - 1) = buf;

    return align_buf;
  }

  return NULL;
}

void rect_free(void *buf)
{
  if (buf != NULL) {
    FreeVec(*(((void **)buf) - 1));
  }
}

#undef FreeVec
#undef AllocVec

#define FreeVec(a...) rect_free(a)
#define AllocVec(a...) rect_malloc(a)

#include "cpu68k.h"
#include "reg68k.h"
#include "mem68k.h"

#include "compile.h"

#define IN_COMPILE_C
#include "common.h"

uint8 *cc_test = NULL;
void *ptr = NULL;

#include "ppc.c"

#undef DO_COMPILE
#define DO_COMPILE

#define COMPILER_BUF (1024*1024*4)

static compiler_t ppc_compiler;
static void *compiler_buf = NULL;

#define POINTERS (65536 / 2)
static void **tables = NULL;
static void **table = NULL;
static int num_tables = 0;

static void dump_go(void) { printf("go\n"); }
static void dump_compile(void) { printf("compiler\n"); }
static void dump_ram(void) { printf("ram\n"); }
static void dump_autovector(void) { printf("autovector\n"); }
static void dump_bad(void) { printf("bad\n"); }
static void dump_exit(void) { printf("exit\n"); }

#define CALL(func) \
  OPCODE = lis(0, ((uint32)&func >> 16)); \
  OPCODE = ori(0, 0, ((uint32)&func & 0xffff)); \
  OPCODE = mtlr(0); \
  OPCODE = blrl();

compiler_t *compile_init(void)
{
  int i, j, table_size;
  t_sr flags;
  void *tmp;
  uint8 cc;

  /* Tables */
  num_tables = 0x40 + 0x01 + 0x01; /* ROM + RAM + BAD */
  table_size = POINTERS * sizeof(void *);

  tables = AllocVec(256 * sizeof(void *), MEMF_CLEAR);
  table = AllocVec(num_tables * table_size, MEMF_CLEAR);
  cc_test = AllocVec(256, MEMF_CLEAR);

  /* Compiler buffer */
  compiler_buf = AllocVec(COMPILER_BUF, MEMF_EXECUTABLE|MEMF_CLEAR);
  ptr = compiler_buf;
  ppc_compiler.first = compiler_buf;
  ppc_compiler.last = ppc_compiler.first + COMPILER_BUF;

  if ((tables == NULL) || (table == NULL) || (cc_test == NULL) || (compiler_buf == NULL)) {
    printf("memory allocation failed\n");
    exit(0);
  }

  /* Go */
  ppc_compiler.go = ptr;
  ppc_save(); /* SAVE all registers */

  CALL(dump_go);

  /* Next Event */
  OPCODE = lis(TABLES, ((uint32)tables >> 16));
  OPCODE = ori(TABLES, TABLES, ((uint32)tables & 0xffff));

  OPCODE = lis(CPUPTR, ((uint32)&cpu68k_clocks >> 16));
  OPCODE = ori(CPUPTR, CPUPTR, ((uint32)&cpu68k_clocks & 0xffff));

  OPCODE = lis(VDPPTR, ((uint32)&vdp_nextevent >> 16));
  OPCODE = ori(VDPPTR, VDPPTR, ((uint32)&vdp_nextevent & 0xffff));

  OPCODE = lis(0, ((uint32)&event_compiler_catchup >> 16));
  OPCODE = ori(0, 0, ((uint32)&event_compiler_catchup & 0xffff));
  OPCODE = mtlr(0);
  OPCODE = blrl();

  OPCODE = lwz(CLKS, 0, VDPPTR);

  OPCODE = rlwinm(6, PC, 18, 22, 29);
  OPCODE = lwzx(7, 6, TABLES);
  OPCODE = rlwinm(6, PC,  1, 15, 29);
  OPCODE = lwzx(7, 6, 7);
  OPCODE = mtlr(7);
  OPCODE = blr();

  /* Exit */
  ppc_compiler.exit = ptr;
  CALL(dump_exit);

  ppc_load(); /* LOAD all registers */
  OPCODE = blr(); /* return to caller */

  /* Compile */
  ppc_compiler.compile = ptr;
//  CALL(dump_compile);

  OPCODE = lis(0, ((uint32)&event_compiler_compile >> 16));
  OPCODE = ori(0, 0, ((uint32)&event_compiler_compile & 0xffff));
  OPCODE = mtlr(0);
  OPCODE = blrl();
  OPCODE = mtlr(3);
  OPCODE = blr();

  /* RAM */
  ppc_compiler.ram = ptr;
//  CALL(dump_ram);
//  OPCODE = b((int)ppc_compiler.exit - (int)ptr);

  OPCODE = stw(CLKS, 0, VDPPTR);

  OPCODE = lis(0, ((uint32)&event_compiler_ram >> 16));
  OPCODE = ori(0, 0, ((uint32)&event_compiler_ram & 0xffff));
  OPCODE = mtlr(0);
  OPCODE = blrl();

  OPCODE = lwz(CLKS, 0, VDPPTR);
  OPCODE = cmpwi(0, CLKS, 0);

  tmp = ptr;
  OPCODE = nop();

  OPCODE = lis(0, ((uint32)&event_compiler_catchup >> 16));
  OPCODE = ori(0, 0, ((uint32)&event_compiler_catchup & 0xffff));
  OPCODE = mtlr(0);
  OPCODE = blrl();

  OPCODE = lwz(CLKS, 0, VDPPTR);

  *((uint32 *)tmp) = bgtL(ptr - tmp); /* likely */

  OPCODE = rlwinm(6, PC, 18, 22, 29);
  OPCODE = lwzx(7, 6, TABLES);
  OPCODE = rlwinm(6, PC,  1, 15, 29);
  OPCODE = lwzx(7, 6, 7);
  OPCODE = mtlr(7);
  OPCODE = blr();

  /* Autovector */
  ppc_compiler.autovector = ptr;
  CALL(dump_autovector);
  OPCODE = b((int)ppc_compiler.exit - (int)ptr);

  /* Bad */
  ppc_compiler.bad = ptr;
  CALL(dump_bad);
  OPCODE = b((int)ppc_compiler.exit - (int)ptr);

  /* Flush cache */
  ptr = ALIGN_PTR(ptr);
  cache_flush(ppc_compiler.go, ((int)ptr - (int)ppc_compiler.go));

  /* Tables */
  for (i=0x00; i<=0xff; i++) { /* BAD */
    tables[i] = (void *)table + (table_size * (0x40 + 0x01));
  }

  for (i=0x00; i<=0x3f; i++) { /* ROM */
    tables[i] = (void *)table + (table_size * i);
  }

  for (i=0xe0; i<=0xff; i++) { /* RAM */
    tables[i] = (void *)table + (table_size * (0x40));
  }

  i=0;
  for (j=0; j<0x40*POINTERS; j++) { /* ROM */
    table[i] = ppc_compiler.compile;
    i++;
  }

  for (j=0; j<0x01*POINTERS; j++) { /* RAM */
    table[i] = ppc_compiler.ram;
    i++;
  }

  for (j=0; j<0x01*POINTERS; j++) { /* BAD */
    table[i] = ppc_compiler.bad;
    i++;
  }

  /* flag positions
   *
   * 000000000000nzvc0000000000000000
   *
   * c = 15
   * v = 14
   * z = 13
   * n = 12
   */
//  *((uint32 *)&flags) = 0x00000000;
  memset(&flags, 0, sizeof(flags));

  for (i=0; i<16; i++) {
    flags.sr_struct.c = (i >> 0) & 1;
    flags.sr_struct.v = (i >> 1) & 1;
    flags.sr_struct.z = (i >> 2) & 1;
    flags.sr_struct.n = (i >> 3) & 1;

    #define NFLAG (flags.sr_struct.n)
    #define ZFLAG (flags.sr_struct.z)
    #define VFLAG (flags.sr_struct.v)
    #define CFLAG (flags.sr_struct.c)

    for (j=0; j<16; j++) {
      switch(j) {
      case 0:                                                     /*  T */
        cc = 1;
        break;
      case 1:                                                     /*  F */
        cc = 0;
        break;
      case 2:                                                     /* HI */
        cc = !(CFLAG || ZFLAG);
        break;
      case 3:                                                     /* LS */
        cc = CFLAG || ZFLAG;
        break;
      case 4:                                                     /* CC */
        cc = !CFLAG;
        break;
      case 5:                                                     /* CS */
        cc = CFLAG;
        break;
      case 6:                                                     /* NE */
        cc = !ZFLAG;
        break;
      case 7:                                                     /* EQ */
        cc = ZFLAG;
        break;
      case 8:                                                     /* VC */
        cc = !VFLAG;
        break;
      case 9:                                                     /* VS */
        cc = VFLAG;
        break;
      case 10:                                                    /* PL */
        cc = !NFLAG;
        break;
      case 11:                                                    /* MI */
        cc = NFLAG;
        break;
      case 12:                                                    /* GE */
        cc = (NFLAG == VFLAG);
        break;
      case 13:                                                    /* LT */
        cc = (NFLAG != VFLAG);
        break;
      case 14:                                                    /* GT */
        cc = !ZFLAG && (NFLAG == VFLAG);
        break;
      case 15:                                                    /* LE */
        cc = ZFLAG || (NFLAG != VFLAG);
        break;
      default:
        cc = 0; /* can never happend, just to remove compiler warning */
        break;
      }

      /* cc_test[(type_of_test << 4) | flags] */
      cc_test[(j << 4) | i] = ((!cc) << 5);
    }
  }

  ppc_compiler.compiled = ptr;

  return &ppc_compiler;
}

void *compile_make(t_ipclist *list)
{
#ifndef DO_COMPILE
  t_ipc *ipc = (t_ipc *)(list+1);
#endif
  void *first = ptr, *tmp;

  /* setup pointer */
  tmp = tables[((list->pc >> 16) & 0xff)];
  ((void **)tmp)[((list->pc >> 1) & 0x7fff)] = ptr;

#ifdef DO_COMPILE
  compile_make_block(list);
#else
  OPCODE = lis(IPC, (((uint32)ipc - sizeof(t_ipc)) >> 16));
  OPCODE = ori(IPC, IPC, (((uint32)ipc - sizeof(t_ipc)) & 0xffff));

  do {
    OPCODE = lwzu(0, sizeof(t_ipc), IPC);
    OPCODE = mr(3, IPC);
    OPCODE = mtlr(0);
    OPCODE = blrl();
    ipc++;
  } while(*(int *)ipc);
#endif

  /* Next Event */
  if (list->norepeat) {
    OPCODE = lwz(7, 0, CPUPTR);

#if 1
    OPCODE = li(6, list->clocks);
    OPCODE = addi(5, CLKS, (list->clocks - 1));
    OPCODE = divw(5, 5, 6);
    OPCODE = mullw(5, 5, 6);

    OPCODE = add(7, 7, 5);
    OPCODE = sub(CLKS, CLKS, 5);
#else
    OPCODE = addi(7, 7, list->clocks);
    OPCODE = subic_(CLKS, CLKS, list->clocks);
    OPCODE = bgtL(-4*2); /* likely */
#endif

    OPCODE = stw(7, 0, CPUPTR);
  } else {
    OPCODE = lwz(7, 0, CPUPTR);
    OPCODE = addi(7, 7, list->clocks);
    OPCODE = stw(7, 0, CPUPTR);

    OPCODE = subic_(CLKS, CLKS, list->clocks);
    tmp = ptr;
    OPCODE = nop();
  }

  OPCODE = stw(CLKS, 0, VDPPTR);

  OPCODE = lis(0, ((uint32)&event_compiler_catchup >> 16));
  OPCODE = ori(0, 0, ((uint32)&event_compiler_catchup & 0xffff));
  OPCODE = mtlr(0);
  OPCODE = blrl();

  OPCODE = lwz(CLKS, 0, VDPPTR);

  if (!list->norepeat) {
    *((uint32 *)tmp) = bgtL(ptr - tmp); /* likely */
  }

  OPCODE = rlwinm(6, PC, 18, 22, 29);
  OPCODE = lwzx(7, 6, TABLES);
  OPCODE = rlwinm(6, PC,  1, 15, 29);
  OPCODE = lwzx(7, 6, 7);
  OPCODE = mtlr(7);
  OPCODE = blr();

  ptr = ALIGN_PTR(ptr);
  cache_flush(first, ((int)ptr - (int)first));

  return first;
}

void compile_quit(void)
{
  int i;

  for (i=0; i<num_tables*POINTERS; i++) {
    table[i] = ppc_compiler.exit;
  }
}

#define FREE(x) if (x != NULL) { FreeVec(x); x = NULL; }

void compile_exit(void)
{
  FREE(compiler_buf);
  FREE(tables);
  FREE(table);
  FREE(cc_test);
}
