/* Generator is (c) James Ponder, 1997-2001 http://www.squish.net/generator/ */

#undef CC_TABLE
//#define CC_TABLE

/* flag positions
 *
 * c = 15
 * v = 14
 * z = 13
 * n = 12
 * x = 11
 */

#ifdef CC_TABLE
void compile_cc(t_iib *iib, int cc_reg, int scale)
{
  uint8 *addr = cc_test + ((iib->cc) << 4);

  if (scale == 5) {
    OPCODE = lis(5, ((uint32)addr >> 16));
    OPCODE = ori(5, 5, ((uint32)addr & 0xffff));
    OPCODE = rlwinm(0, SRREG, 16, 28, 31);
    OPCODE = lbzx(cc_reg, 5, 0);
  } else {
    printf("CC: Error\n");
    instr_error = 1;
  }
}
#else
void compile_cc(t_iib *iib, int cc_reg, int scale)
{
  switch(iib->cc) {
  case  0: /*  T = 1 */
    OPCODE = li(cc_reg, 1 << scale);
    break;
  case  1: /*  F = 0 */
    OPCODE = li(cc_reg, 0 << scale);
    break;
  case  2: /* HI = ~(C | Z) */
    OPCODE = rlwinm(0, SRREG, 30, 0, 31);
    OPCODE = nor(0, 0, SRREG);
    OPCODE = rlwinm(cc_reg, 0, 16 + scale, 31 - scale, 31 - scale);
    break;
  case  3: /* LS = (C | Z) */
    OPCODE = rlwinm(0, SRREG, 30, 0, 31);
    OPCODE = or(0, 0, SRREG);
    OPCODE = rlwinm(cc_reg, 0, 16 + scale, 31 - scale, 31 - scale);
    break;
  case  4: /* CC = ~C */
    OPCODE = not(cc_reg, SRREG);
    OPCODE = rlwinm(cc_reg, cc_reg, 16 + scale, 31 - scale, 31 - scale);
    break;
  case  5: /* CS = C */
    OPCODE = rlwinm(cc_reg, SRREG, 16 + scale, 31 - scale, 31 - scale);
    break;
  case  6: /* NE = ~Z */
    OPCODE = not(cc_reg, SRREG);
    OPCODE = rlwinm(cc_reg, cc_reg, 14 + scale, 31 - scale, 31 - scale);
    break;
  case  7: /* EQ = Z */
    OPCODE = rlwinm(cc_reg, SRREG, 14 + scale, 31 - scale, 31 - scale);
    break;
  case  8: /* VC = ~V */
    OPCODE = not(cc_reg, SRREG);
    OPCODE = rlwinm(cc_reg, cc_reg, 15 + scale, 31 - scale, 31 - scale);
    break;
  case  9: /* VS = V */
    OPCODE = rlwinm(cc_reg, SRREG, 15 + scale, 31 - scale, 31 - scale);
    break;
  case 10: /* PL = ~N */
    OPCODE = not(cc_reg, SRREG);
    OPCODE = rlwinm(cc_reg, cc_reg, 13 + scale, 31 - scale, 31 - scale);
    break;
  case 11: /* MI = N */
    OPCODE = rlwinm(cc_reg, SRREG, 13 + scale, 31 - scale, 31 - scale);
    break;
  case 12: /* GE = ~(N ^ V) */
    OPCODE = rlwinm(0, SRREG, 30, 0, 31);
    OPCODE = eqv(0, 0, SRREG);
    OPCODE = rlwinm(cc_reg, 0, 15 + scale, 31 - scale, 31 - scale);
    break;
  case 13: /* LT = (N ^ V) */
    OPCODE = rlwinm(0, SRREG, 30, 0, 31);
    OPCODE = xor(0, 0, SRREG);
    OPCODE = rlwinm(cc_reg, 0, 15 + scale, 31 - scale, 31 - scale);
    break;
  case 14: /* GT = ~((N ^ V) | Z) */
    OPCODE = rlwinm(0, SRREG, 30, 0, 31);
    OPCODE = xor(cc_reg, 0, SRREG);
    OPCODE = rlwinm(0, SRREG, 31, 0, 31);
    OPCODE = nor(cc_reg, cc_reg, 0);
    OPCODE = rlwinm(cc_reg, cc_reg, 15 + scale, 31 - scale, 31 - scale);
    break;
  case 15: /* LE = (N ^ V) | Z */
    OPCODE = rlwinm(0, SRREG, 30, 0, 31);
    OPCODE = xor(cc_reg, 0, SRREG);
    OPCODE = rlwinm(0, SRREG, 31, 0, 31);
    OPCODE = or(cc_reg, cc_reg, 0);
    OPCODE = rlwinm(cc_reg, cc_reg, 15 + scale, 31 - scale, 31 - scale);
    break;
  default:
    printf("CC: Error\n");
    instr_error = 1;
    break;
  }
}
#endif

/* clrflag */

void compile_clrflag_c(t_iib *iib)
{
  OPCODE = rlwinm(SRREG, SRREG, 0, 16, 14); /* clear bit 15 */
}

void compile_clrflag_v(t_iib *iib)
{
  OPCODE = rlwinm(SRREG, SRREG, 0, 15, 13); /* clear bit 14 */
}

void compile_clrflag_z(t_iib *iib)
{
  OPCODE = rlwinm(SRREG, SRREG, 0, 14, 12); /* clear bit 13 */
}

void compile_clrflag_n(t_iib *iib)
{
  OPCODE = rlwinm(SRREG, SRREG, 0, 13, 11); /* clear bit 12 */
}

void compile_clrflag_x(t_iib *iib)
{
  OPCODE = rlwinm(SRREG, SRREG, 0, 12, 10); /* clear bit 11 */
}

void compile_clrflags(uint32 set)
{
  uint32 mask = 0;

  if (set & IIB_FLAG_C) mask |= (1 << 16);
  if (set & IIB_FLAG_V) mask |= (1 << 17);
  if (set & IIB_FLAG_Z) mask |= (1 << 18);
  if (set & IIB_FLAG_N) mask |= (1 << 19);
  if (set & IIB_FLAG_X) mask |= (1 << 20);

  mask = ~mask;

  OPCODE = andis_(SRREG, SRREG, (mask >> 16));
}

/* setflag */

void compile_setflag_c(t_iib *iib)
{
  OPCODE = oris(SRREG, SRREG, (1 << (15 - 15))); /* set bit 15 */
}

void compile_setflag_v(t_iib *iib)
{
  OPCODE = oris(SRREG, SRREG, (1 << (15 - 14))); /* set bit 14 */
}

void compile_setflag_z(t_iib *iib)
{
  OPCODE = oris(SRREG, SRREG, (1 << (15 - 13))); /* set bit 13 */
}

void compile_setflag_n(t_iib *iib)
{
  OPCODE = oris(SRREG, SRREG, (1 << (15 - 12))); /* set bit 12 */
}

void compile_setflag_x(t_iib *iib)
{
  OPCODE = oris(SRREG, SRREG, (1 << (15 - 11))); /* set bit 11 */
}

void compile_setflags(uint32 set)
{
  uint32 mask = 0;

  if (set & IIB_FLAG_C) mask |= (1 << 16);
  if (set & IIB_FLAG_V) mask |= (1 << 17);
  if (set & IIB_FLAG_Z) mask |= (1 << 18);
  if (set & IIB_FLAG_N) mask |= (1 << 19);
  if (set & IIB_FLAG_X) mask |= (1 << 20);

  OPCODE = oris(SRREG, SRREG, (mask >> 16));
}

/* stdflag */

void compile_stdflag_n(t_iib *iib, int reg_out)
{
  /* reg_out >> 31 */

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwimi(SRREG, reg_out, 12, 12, 12);
    break;
  case sz_word:
    OPCODE = rlwimi(SRREG, reg_out,  4, 12, 12);
    break;
  case sz_long:
    OPCODE = rlwimi(SRREG, reg_out, 20, 12, 12);
    break;
  default:
    printf("STDFLAG_N: Invalid size %d\n", iib->size);
    instr_error = 1;
    break;
  }
}

void compile_stdflag_z(t_iib *iib, int reg_out)
{
  /* reg_out == 31 */

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwinm(0, reg_out, 0, 24, 31);
    OPCODE = cntlzw(0, 0); /* bit 26 = (reg_out == 0) */
    break;
  case sz_word:
    OPCODE = rlwinm(0, reg_out, 0, 16, 31);
    OPCODE = cntlzw(0, 0); /* bit 26 = (reg_out == 0) */
    break;
  case sz_long:
    OPCODE = cntlzw(0, reg_out); /* bit 26 = (reg_out == 0) */
    break;
  default:
    printf("STDFLAG_Z: Invalid size %d\n", iib->size);
    instr_error = 1;
    break;
  }

  OPCODE = rlwimi(SRREG, 0, 13, 13, 13); /* ZFLAG */
}

void compile_stdxflag_z(t_iib *iib, int reg_out)
{
  /* reg_out == 31 */

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwinm(0, reg_out, 0, 24, 31);
    OPCODE = cntlzw(0, 0); /* bit 26 = (reg_out == 0) */
    break;
  case sz_word:
    OPCODE = rlwinm(0, reg_out, 0, 16, 31);
    OPCODE = cntlzw(0, 0); /* bit 26 = (reg_out == 0) */
    break;
  case sz_long:
    OPCODE = cntlzw(0, reg_out); /* bit 26 = (reg_out == 0) */
    break;
  default:
    printf("STDFLAG_Z: Invalid size %d\n", iib->size);
    instr_error = 1;
    break;
  }

  OPCODE = rlwinm(0, 0, 13, 13, 13); /* ZFLAG */
  OPCODE = xoris(0, 0, 0x0004);
  OPCODE = andc(SRREG, SRREG, 0);
}

/* addflag (ADD, ADDX) */

void compile_addflag_v(t_iib *iib, int reg_src, int reg_dst, int reg_out)
{
  /* (reg_src ^ reg_out) & (reg_dst ^ reg_out) */

  OPCODE = xor(0, reg_src, reg_out);
  OPCODE = xor(5, reg_dst, reg_out);
  OPCODE = and(0, 0, 5);

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwimi(SRREG, 0, 10, 14, 14);
    break;
  case sz_word:
    OPCODE = rlwimi(SRREG, 0,  2, 14, 14);
    break;
  case sz_long:
    OPCODE = rlwimi(SRREG, 0, 18, 14, 14);
    break;
  default:
    printf("ADDFLAG_V: Invalid size %d\n", iib->size);
    instr_error = 1;
    break;
  }
}

/* (ADD) */

void compile_addflag_cx(t_iib *iib, int reg_src, int reg_dst, int reg_out)
{
  switch(iib->size) {
  case sz_byte: /* & 0x00000100 */
    OPCODE = rlwimi(SRREG, reg_out,  8, 15, 15);
    OPCODE = rlwimi(SRREG, reg_out, 12, 11, 11);
    break;
  case sz_word: /* & 0x00010000 */
    OPCODE = rlwimi(SRREG, reg_out,  0, 15, 15);
    OPCODE = rlwimi(SRREG, reg_out,  4, 11, 11);
    break;
  case sz_long: /* FIXME */
//    OPCODE = addc(0, reg_src, reg_dst);
    OPCODE = li(0, 0);
    OPCODE = addze(0, 0);
    OPCODE = rlwimi(SRREG, 0, 16, 15, 15);
    OPCODE = rlwimi(SRREG, 0, 20, 11, 11);
    break;
  default:
    printf("ADDFLAG_CX: Invalid size %d\n", iib->size);
    instr_error = 1;
    break;
  }
}

/* subflag (SUB, SUBX, CMP) */

void compile_subflag_v(t_iib *iib, int reg_src, int reg_dst, int reg_out)
{
  /* (reg_src ^ reg_dst) & (reg_out ^ reg_dst) */

  OPCODE = xor(0, reg_src, reg_dst);
  OPCODE = xor(5, reg_out, reg_dst);
  OPCODE = and(0, 0, 5);

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwimi(SRREG, 0, 10, 14, 14);
    break;
  case sz_word:
    OPCODE = rlwimi(SRREG, 0,  2, 14, 14);
    break;
  case sz_long:
    OPCODE = rlwimi(SRREG, 0, 18, 14, 14);
    break;
  default:
    printf("SUBFLAG_V: Invalid size %d\n", iib->size);
    instr_error = 1;
    break;
  }
}

/* (CMP) */

void compile_subflag_c(t_iib *iib, int reg_src, int reg_dst, int reg_out)
{
  /* (reg_dst - reg_src) >> 31 */

  if (iib->size == sz_long) { /* FIXME */
//    OPCODE = subc(0, reg_dst, reg_src);
    OPCODE = li(0, 1);
    OPCODE = addze(0, 0);
    OPCODE = rlwimi(SRREG, 0, 16, 15, 15); /* CFLAG */
  } else {
    OPCODE = rlwimi(SRREG, reg_out, 17, 15, 15); /* CFLAG */
  }
}

/* (SUB) */

void compile_subflag_cx(t_iib *iib, int reg_src, int reg_dst, int reg_out)
{
  /* (reg_dst - reg_src) >> 31 */

  if (iib->size == sz_long) { /* FIXME */
//    OPCODE = subc(0, reg_dst, reg_src);
    OPCODE = li(0, 1);
    OPCODE = addze(0, 0);
    OPCODE = rlwimi(SRREG, 0, 16, 15, 15); /* CFLAG */
    OPCODE = rlwimi(SRREG, 0, 20, 11, 11); /* XFLAG */
  } else {
    OPCODE = rlwimi(SRREG, reg_out, 17, 15, 15); /* CFLAG */
    OPCODE = rlwimi(SRREG, reg_out, 21, 11, 11); /* XFLAG */
  }
}

/* negflag (NEG) */

void compile_negflag_v(t_iib *iib, int reg_src, int reg_out)
{
  /* overflow if MSB in src AND out is set */

  OPCODE = and(0, reg_src, reg_out);

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwimi(SRREG, 0, 10, 14, 14);
    break;
  case sz_word:
    OPCODE = rlwimi(SRREG, 0,  2, 14, 14);
    break;
  case sz_long:
    OPCODE = rlwimi(SRREG, 0, 18, 14, 14);
    break;
  default:
    printf("SUBFLAG_V: Invalid size %d\n", iib->size);
    instr_error = 1;
    break;
  }
}

/* (NEG) */

void compile_negflag_cx(t_iib *iib, int reg_out)
{
  OPCODE = cntlzw(0, reg_out); /* bit 26 = (reg_out == 0) */
  OPCODE = xori(0, 0, 32);
  OPCODE = rlwimi(SRREG, 0, 11, 15, 15); /* CFLAG */
  OPCODE = rlwimi(SRREG, 0, 15, 11, 11); /* XFLAG */
}

/* negxflag (NEGX) */

void compile_negxflag_v(t_iib *iib, int reg_src, int reg_out)
{
  /* overflow if MSB in src AND dst is set */

  OPCODE = and(0, reg_src, reg_out);

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwimi(SRREG, 0, 10, 14, 14);
    break;
  case sz_word:
    OPCODE = rlwimi(SRREG, 0,  2, 14, 14);
    break;
  case sz_long:
    OPCODE = rlwimi(SRREG, 0, 18, 14, 14);
    break;
  default:
    printf("NEGXFLAG_V: Invalid size %d\n", iib->size);
    instr_error = 1;
    break;
  }
}

/* (NEGX) */

void compile_negxflag_cx(t_iib *iib, int reg_src, int reg_out)
{
  /* overflow if MSB in src OR dst is set */

  OPCODE = or(0, reg_src, reg_out);

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwimi(SRREG, 0,  9, 15, 15);
    OPCODE = rlwimi(SRREG, 0, 13, 11, 11);
    break;
  case sz_word:
    OPCODE = rlwimi(SRREG, 0,  1, 15, 15);
    OPCODE = rlwimi(SRREG, 0,  5, 11, 11);
    break;
  case sz_long:
    OPCODE = rlwimi(SRREG, 0, 17, 15, 15);
    OPCODE = rlwimi(SRREG, 0, 21, 11, 11);
    break;
  default:
    printf("NEGXFLAG_V: Invalid size %d\n", iib->size);
    instr_error = 1;
    break;
  }
}
