/* Generator is (c) James Ponder, 1997-2001 http://www.squish.net/generator/ */

/* POSSIBLE OPTIMIZATIONS
 *
 * - Maybe some PC-optimizations can be made now since the "pc_inc" flag is there.
 * - It may be possible to use PPC processor flags at times.
 * - Implement a static and/or dynamic register cache.
 * - Merge flag calculation with opcode operation to get advantages.
 */

/* KNOWN BUGS
 *
 * - Undefined flags may not be calculated the way test show they are calculated.
 * - Overrun of the compiler buffer will crash the emulator.
 * - Many of the flag calculations are not correct for edge values (min/max).
 * - MOVE.L with pre-decrementing destination is probably handled wrong.
 */

#undef DEBUG
//#define DEBUG

#define DATAREG(a) ((a) * 4)
#define ADDRREG(a) ((8+(a)) * 4)

#include "generator.h"
#include "registers.h"

#include "cpu68k.h"
#include "reg68k.h"
#include "mem68k.h"

#include "ppc.h"
#include "common.h"

#include <stdio.h>
#include <string.h>

#if 0

#define LI(reg, val) {                                 \
  sint32 imm = val;                                    \
  if ((imm >= -32768) && (imm <= 32767)) {             \
    OPCODE = li(reg, ((uint32)imm & 0xffff));          \
  } else {                                             \
    OPCODE = lis(reg, ((uint32)imm >> 16));            \
    OPCODE = ori(reg, reg, ((uint32)imm & 0xffff));    \
  }                                                    \
}

static unsigned short imm_h, imm_l;
static unsigned short add_h, add_l;
static unsigned short sub_h, sub_l;

void SI(unsigned long val)
{
  imm_h = (unsigned short)(val >> 16);
  imm_l = (unsigned short)(val & 0xffff);

  add_h = (unsigned short)((val + 32768) >> 16);
  add_l = (unsigned short)(val & 0xffff);

  val = 0 - val;

  sub_h = (unsigned short)((val + 32768) >> 16);
  sub_l = (unsigned short)(val & 0xffff);
}

/* ADD : addis 0, 0, add_h; addi 0, 0, add_l; */
/* SUB : addis 0, 0, sub_h; addi 0, 0, sub_l; */
/*  OR :  oris 0, 0, imm_h;  ori 0, 0, imm_l; */

#endif

static int instr_error; /* When set, compiling failed, use interpreter instr */

static int current_work_reg = 12;

static int get_work_reg(int force)
{
  int ret = current_work_reg;

  /* Allow us to force a register */
  if (force > 0) {
    return force;
  }

  current_work_reg--;

  if ((ret < 7) || (ret > 12)) {
    printf("ERROR: out of workregs? (%d)\n", ret);
    current_work_reg = 12;
    ret = 12;
  }

/*
  printf("work reg: %d\n", ret);
*/

  return ret;
}

static void reset_work_reg(void)
{
  current_work_reg = 12;
}

#include "compile_flags.h" /*** flags ***/

static void compile_ea(t_iib *iib, t_ipc *ipc, t_type type, int update, int *reg, int force)
{
  t_datatype datatype = type ? iib->dtype : iib->stype;
  uint32 imm, srcdst;
  int reg68, idxreg;
  int a, size;

  /* compile EA to register(s) regea/regval */

  reg68 = 8 + ((ipc->opcode >> (type ? iib->dbitpos : iib->sbitpos)) & 7);
  reg68 = (reg68 * 4) + 0;

  switch(datatype) {
  case dt_Dreg:
  case dt_Areg:
    *reg = -1;
    break;
  case dt_Aind:
    a = get_work_reg(force);
    OPCODE = lwz(a, reg68, DAREG);
    *reg = a;
    break;
  case dt_Ainc:
    a = get_work_reg(force);
    OPCODE = lwz(a, reg68, DAREG);
    if (update) {
      if (iib->size == sz_byte) {
        size = ((reg68 - 16 - 8) == 7 ? 2 : 1);
      } else {
        size = (1<<(iib->size-1));
      }
      OPCODE = addi(0, a, size);
      OPCODE = stw(0, reg68, DAREG);
    }
    *reg = a;
    break;
  case dt_Adec:
    a = get_work_reg(force);
    OPCODE = lwz(a, reg68, DAREG);
    if (update) {
      if (iib->size == sz_byte) {
        size = ((reg68 - 16 - 8) == 7 ? 2 : 1);
      } else {
        size = (1<<(iib->size-1));
      }
      OPCODE = subi(a, a, size);
      OPCODE = stw(a, reg68, DAREG);
    }
    *reg = a;
    break;
  case dt_Adis:
    a = get_work_reg(force);
    OPCODE = lwz(a, reg68, DAREG);

    imm = (type ? ipc->dst : ipc->src);

    OPCODE = addi(a, a, imm);
    *reg = a;
    break;
  case dt_Aidx:
    srcdst = (type ? ipc->dst : ipc->src);

    idxreg = ((srcdst>>28) & 15); /* idx reg */
    idxreg = (idxreg * 4) + 0;

    imm = ((((sint32)(srcdst<<8)))>>8);
    OPCODE = lis(0, ((uint32)imm >> 16));
    OPCODE = ori(0, 0, ((uint32)imm & 0xffff));

    a = get_work_reg(force);
    if ((srcdst>>27) & 1) {
      OPCODE = lwz(a, (idxreg + 0), DAREG);
    } else {
      OPCODE = lha(a, (idxreg + 2), DAREG);
    }
    OPCODE = add(a, a, 0);

    /* Add "A" register */
    OPCODE = lwz(0, reg68, DAREG);
    OPCODE = add(a, a, 0);

    *reg = a;
    break;
  case dt_AbsW:
    a = get_work_reg(force);

    imm = (type ? ipc->dst : ipc->src);

    OPCODE = li(a, ((uint32)imm & 0xffff));
    *reg = a;
    break;
  case dt_AbsL:
  case dt_Pdis:
    a = get_work_reg(force);

    imm = (type ? ipc->dst : ipc->src);

    OPCODE = lis(a, ((uint32)imm >> 16));
    OPCODE = ori(a, a, ((uint32)imm & 0xffff));
    *reg = a;
    break;
  case dt_Pidx:
    srcdst = (type ? ipc->dst : ipc->src);

    idxreg = ((srcdst>>28) & 15); /* idx reg */
    idxreg = (idxreg * 4) + 0;

    imm = ((((sint32)(srcdst<<8)))>>8);
    OPCODE = lis(0, ((uint32)imm >> 16));
    OPCODE = ori(0, 0, ((uint32)imm & 0xffff));

    a = get_work_reg(force);
    if ((srcdst>>27) & 1) {
      OPCODE = lwz(a, (idxreg + 0), DAREG);
    } else {
      OPCODE = lha(a, (idxreg + 2), DAREG);
    }
    OPCODE = add(a, a, 0);

    *reg = a;
    break;
  case dt_ImmB:
  case dt_ImmW:
  case dt_ImmL:
  case dt_ImmS:
  case dt_Imm3:
  case dt_Imm4:
  case dt_Imm8:
  case dt_Imm8s:
    *reg = -1;
    break;
  default:
    printf("EA: Invalid datatype %d\n", datatype);
    instr_error = 1;
  }
}

static void compile_eaval(t_iib *iib, t_ipc *ipc, t_type type, int eareg, int *eaval, int force)
{
  t_datatype datatype = type ? iib->dtype : iib->stype;
  uint32 imm, fetch_function = 0;
  int a, reg68;

  /* compile EA value */

  switch(datatype) {
  case dt_Dreg:
  case dt_Areg:
    reg68 = ((datatype == dt_Areg ? 8 : 0) + ((ipc->opcode >> (type ? iib->dbitpos : iib->sbitpos)) & 7));
    a = get_work_reg(force);

    switch(iib->size) {
    case sz_byte:
      reg68 = (reg68 * 4) + 3;
      OPCODE = lbz(a, reg68, DAREG);
      break;
    case sz_word:
      reg68 = (reg68 * 4) + 2;
      OPCODE = lhz(a, reg68, DAREG);
      break;
    case sz_long:
      reg68 = (reg68 * 4) + 0;
      OPCODE = lwz(a, reg68, DAREG);
      break;
    default:
      printf("EAVAL: Invalid size %d\n", iib->size);
      instr_error = 1;
    }

    *eaval = a;
    break;
  case dt_Aind:
  case dt_Ainc:
  case dt_Adec:
  case dt_Adis:
  case dt_Aidx:
  case dt_AbsW:
  case dt_AbsL:
  case dt_Pdis:
  case dt_Pidx:
    switch (iib->size) {
    case sz_byte:
      fetch_function = (uint32)mem68k_fetch_byte;
      break;
    case sz_word:
      fetch_function = (uint32)mem68k_fetch_word;
      break;
    case sz_long:
      fetch_function = (uint32)mem68k_fetch_long;
      break;
    default:
      printf("EAVAL: Invalid size %d\n", iib->size);
      instr_error = 1;
    }

    OPCODE = rlwinm(3, eareg, 0, 8, 31);
    OPCODE = lis(6, ((uint32)fetch_function >> 16));
    OPCODE = ori(6, 6, ((uint32)fetch_function & 0xffff));
    OPCODE = rlwinm(0, 3, 22, 18, 29);
    OPCODE = lwzx(0, 6, 0);
    OPCODE = mtlr(0);
    OPCODE = blrl();

    /* Store away reg 3 somewhere else, required when using two EAVAL (ADD for example) */
    a = get_work_reg(force);
    OPCODE = mr(a, 3);
    *eaval = a;
    break;
  case dt_ImmB:  /* 8bit unsigned */
  case dt_Imm3:  /* 3bit unsigned */
  case dt_Imm4:  /* 4bit unsigned */
  case dt_Imm8:  /* 8bit unsigned */
  case dt_Imm8s: /* 8bit signed */
    a = get_work_reg(force);

    imm = (type ? ipc->dst : ipc->src);

    OPCODE = li(a, ((uint32)imm & 0xffff));
    *eaval = a;
    break;
  case dt_ImmW: /* 16bit unsigned */
  case dt_ImmL: /* 32bit unsigned */
    a = get_work_reg(force);

    imm = (type ? ipc->dst : ipc->src);

    OPCODE = lis(a, ((uint32)imm >> 16));
    OPCODE = ori(a, a, ((uint32)imm & 0xffff));
    *eaval = a;
    break;
  case dt_ImmS: /* value is 1 or 8 */
    a = get_work_reg(force);

    imm = iib->immvalue;

    OPCODE = li(a, ((uint32)imm & 0xffff));
    *eaval = a;
    break;
  default:
    printf("EAVAL: Invalid datatype %d\n", datatype);
    instr_error = 1;
  }
}

static void compile_eastore(t_iib *iib, t_ipc *ipc, t_type type, int eareg, int eaval)
{
  t_datatype datatype = type ? iib->dtype : iib->stype;
  uint32 store_function = 0;
  int reg68;

  /* store EA value */

  switch(datatype) {
  case dt_Dreg:
  case dt_Areg:
    reg68 = ((datatype == dt_Areg ? 8 : 0) + ((ipc->opcode >> (type ? iib->dbitpos : iib->sbitpos)) & 7));

    switch(iib->size) {
    case sz_byte:
      reg68 = (reg68 * 4) + 3;
      OPCODE = stb(eaval, reg68, DAREG);
      break;
    case sz_word:
      reg68 = (reg68 * 4) + 2;
      OPCODE = sth(eaval, reg68, DAREG);
      break;
    case sz_long:
      reg68 = (reg68 * 4) + 0;
      OPCODE = stw(eaval, reg68, DAREG);
      break;
    default:
      printf("EASTORE: Invalid size %d\n", iib->size);
      instr_error = 1;
    }

    break;
  case dt_Aind:
  case dt_Ainc:
  case dt_Adec:
  case dt_Adis:
  case dt_Aidx:
  case dt_AbsW:
  case dt_AbsL:
  case dt_Pdis:
  case dt_Pidx:
    switch (iib->size) {
    case sz_byte:
      store_function = (uint32)mem68k_store_byte;
      break;
    case sz_word:
      store_function = (uint32)mem68k_store_word;
      break;
    case sz_long:
      store_function = (uint32)mem68k_store_long;
      break;
    default:
      printf("EASTORE: Invalid size %d\n", iib->size);
      instr_error = 1;
    }

    /* Copy data to register 4 when required */
    if (eaval != 4) {
      OPCODE = mr(4, eaval);
    }
    OPCODE = rlwinm(3, eareg, 0, 8, 31);
    OPCODE = lis(6, ((uint32)store_function >> 16));
    OPCODE = ori(6, 6, ((uint32)store_function & 0xffff));
    OPCODE = rlwinm(0, 3, 22, 18, 29);
    OPCODE = lwzx(0, 6, 0);
    OPCODE = mtlr(0);
    OPCODE = blrl();
    break;
  case dt_ImmB:
  case dt_ImmW:
  case dt_ImmL:
  case dt_Imm3:
  case dt_Imm4:
  case dt_Imm8:
  case dt_Imm8s:
  case dt_ImmS:
  default:
    printf("EASTORE: Invalid datatype %d\n", datatype);
    instr_error = 1;
  }
}

/* Get the destination ADDRESS or DATA register */

static int compile_adreg(t_iib *iib, t_ipc *ipc, t_type type)
{
  t_datatype datatype = type ? iib->dtype : iib->stype;

  switch(datatype) {
  case dt_Dreg:
  case dt_Areg:
    return ((datatype == dt_Areg ? 8 : 0) + ((ipc->opcode >> (type ? iib->dbitpos : iib->sbitpos)) & 7));
    break;
  default:
    return 0;
    break;
  }
}

static void compile_adval(t_iib *iib, t_ipc *ipc, t_type type, int *eaval, int force)
{
  int a, reg68;

  a = 0;
  reg68 = compile_adreg(iib, ipc, type);
  reg68 = (reg68 * 4) + 0;
  OPCODE = lwz(a, reg68, DAREG);

  *eaval = a;
}

static void compile_adstore(t_iib *iib, t_ipc *ipc, t_type type, int eaval)
{
  int reg68;

  reg68 = compile_adreg(iib, ipc, type);
  reg68 = (reg68 * 4) + 0;
  OPCODE = stw(eaval, reg68, DAREG);
}

#ifdef DEBUG

/* forward references */

void diss68k_getoperand(char *text, t_ipc * ipc, t_iib * iib, t_type type);

/* functions */

int diss68k_gettext(t_ipc * ipc, char *text)
{
  t_iib *iib;
  char *p, *c;
  char src[64], dst[64];
  char mnemonic[64];

  *text = '\0';

  iib = cpu68k_iibtable[ipc->opcode];

  if (iib == NULL)
    return 0;

  diss68k_getoperand(dst, ipc, iib, tp_dst);
  diss68k_getoperand(src, ipc, iib, tp_src);

  if ((iib->mnemonic == i_Bcc) || (iib->mnemonic == i_BSR) ||
      (iib->mnemonic == i_DBcc)) {
    sprintf(src, "$%08x", ipc->src);
  }

  strcpy(mnemonic, mnemonic_table[iib->mnemonic].name);

  if ((p = strstr(mnemonic, "cc")) != NULL) {
    if (iib->mnemonic == i_Bcc && iib->cc == 0) {
      p[0] = 'R';
      p[1] = 'A';
    } else {
      c = condition_table[iib->cc];
      strcpy(p, c);
    }
  }

  switch (iib->size) {
  case sz_byte:
    strcat(mnemonic, ".B");
    break;
  case sz_word:
    strcat(mnemonic, ".W");
    break;
  case sz_long:
    strcat(mnemonic, ".L");
    break;
  default:
    break;
  }

  sprintf(text, "%-10s %s%s%s", mnemonic, src, dst[0] ? "," : "", dst);

  return 1;
}

void diss68k_getoperand(char *text, t_ipc * ipc, t_iib * iib, t_type type)
{
  int bitpos;
  uint32 val;

  if (type == tp_src) {
    bitpos = iib->sbitpos;
    val = ipc->src;
  } else {
    bitpos = iib->dbitpos;
    val = ipc->dst;
  }

  switch (type == tp_src ? iib->stype : iib->dtype) {
  case dt_Dreg:
    sprintf(text, "D%d", (ipc->opcode >> bitpos) & 7);
    break;
  case dt_Areg:
    sprintf(text, "A%d", (ipc->opcode >> bitpos) & 7);
    break;
  case dt_Aind:
    sprintf(text, "(A%d)", (ipc->opcode >> bitpos) & 7);
    break;
  case dt_Ainc:
    sprintf(text, "(A%d)+", (ipc->opcode >> bitpos) & 7);
    break;
  case dt_Adec:
    sprintf(text, "-(A%d)", (ipc->opcode >> bitpos) & 7);
    break;
  case dt_Adis:
    sprintf(text, "$%04x(A%d)", (uint16)val, (ipc->opcode >> bitpos) & 7);
    break;
  case dt_Aidx:
    sprintf(text, "$%02x(A%d,Rx.X)", (uint8)val, (ipc->opcode >> bitpos) & 7);
    break;
  case dt_AbsW:
    sprintf(text, "$%08x", val);
    break;
  case dt_AbsL:
    sprintf(text, "$%08x", val);
    break;
  case dt_Pdis:
    sprintf(text, "$%08x(pc)", val);
    break;
  case dt_Pidx:
    sprintf(text, "$%08x(pc, Rx.X)", val);
    break;
  case dt_ImmB:
    sprintf(text, "#$%02x", val);
    break;
  case dt_ImmW:
    sprintf(text, "#$%04x", val);
    break;
  case dt_ImmL:
    sprintf(text, "#$%08x", val);
    break;
  case dt_ImmS:
    sprintf(text, "#%d", iib->immvalue);
    break;
  case dt_Imm3:
    sprintf(text, "#%d", (ipc->opcode >> bitpos) & 7);
    break;
  case dt_Imm4:
    sprintf(text, "#%d", (ipc->opcode >> bitpos) & 15);
    break;
  case dt_Imm8:
    sprintf(text, "#%d", (ipc->opcode >> bitpos) & 255);
    break;
  case dt_Imm8s:
    sprintf(text, "#%d", (sint32)val);
    break;
  default:
    break;
  }
}

static char diss_68k[1024];
#endif

void compile_make_block(t_ipclist *list)
{
  uint32 addr68k;
  void *tmp, *end, *backup;
  t_ipc *ipc = (t_ipc *)(list+1);
  t_iib *iib;
  int normal;
  int srcreg, dstreg, outreg, srcval, dstval, outval;
  int pc_inc;

  addr68k = list->pc;

#ifdef DEBUG
printf("New block @ %08x\n", addr68k);
#endif

  do {
    iib = cpu68k_iibtable[ipc->opcode];
    normal = 0;

    reset_work_reg(); // NEW instruction...
    backup = ptr; // FIXME: Remove when everything is finished
    instr_error = 0;
    pc_inc = 1;

    switch(iib->mnemonic) {
      #include "compile_op_1.h"
      #include "compile_op_2.h"
      #include "compile_op_3.h"
      #include "compile_op_4.h"
      #include "compile_op_5.h"
      #include "compile_op_6.h"
      #include "compile_op_x.h"

      default:
        instr_error = 1; /* Oops */
        break;
    }

    /* Handle various errors... */
    if (instr_error == 1) {
#ifdef DEBUG
      printf("comp failed, fixing...\n");
#endif
      ptr = backup;
      normal = 1;
    }

#ifdef DEBUG
    if (normal == 1) {
      diss68k_gettext(ipc, diss_68k);
      printf("  %s\n", diss_68k);
    }
#endif

    /*Increase PC */
    if ((normal == 0) && (pc_inc == 1)) {
      OPCODE = addi(PCREG, PCREG, 2*iib->wordlen);
    }

    addr68k += 2*iib->wordlen;

    if (normal == 1) {
      OPCODE = lis(3, ((uint32)ipc >> 16));
      OPCODE = ori(3, 3, ((uint32)ipc & 0xffff));
      OPCODE = lis(6, ((uint32)ipc->function >> 16));
      OPCODE = ori(6, 6, ((uint32)ipc->function & 0xffff));
      OPCODE = mtlr(6);
      OPCODE = blrl();
    }
    ipc++;
  } while(*(int *)ipc);
}
