/*****************************************************************************/
/*     Generator - Sega Genesis emulation - (c) James Ponder 1997-2001       */
/*****************************************************************************/
/*                                                                           */
/* cli-amiga.c - Amiga CLI interface (c) Mathias "AmiDog" Roslund 2001       */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "config.h"
#include "generator.h"
#include "ui-console.h"
#include "gensound.h"
#include "cpuz80.h"
#include "cli-amiga.h"

#define OLD_TYPEDEFS
typedef float float32;
typedef double float64;
#include <proto/dos.h>

cli_amiga_type cli_amiga;

const char ver[] = "$VER: AmiGenerator v0.35 (051107)";

#define TEMPLATE "FILE/A,SIMPLE/S,FRAMESKIP/N,NOSOUND/S,NOZ80/S,NOFM/S,NOPSG/S,LACE/K,LATENCY/N,"\
                 "FRAGMENTS/N,JOYPAD/S,FREQ/N,REQ/S,NOSTALGIA/S,INTERPRETER/S,COMPILER/S," \
                 "SPEEDTEST/K"

#define OPT_FILE         0
#define OPT_SIMPLE       1
#define OPT_FRAMESKIP    2
#define OPT_NOSOUND      3
#define OPT_NOZ80        4
#define OPT_NOFM         5
#define OPT_NOPSG        6
#define OPT_LACE         7
#define OPT_LATENCY      8
#define OPT_FRAGMENTS    9
#define OPT_JOYPAD      10
#define OPT_FREQ        11
#define OPT_REQ         12
#define OPT_NOSTALGIA   13
#define OPT_INTERPRETER 14
#define OPT_COMPILER    15
#define OPT_SPEEDTEST   16

#define OPT_COUNT       17

void handle_arguments(int argc, char *argv[])
{
  struct RDArgs *rdargs = NULL;
  long opts[OPT_COUNT];

  memset(opts, 0, sizeof(opts));
  memset(&cli_amiga, 0, sizeof(cli_amiga_type));

  if ((rdargs = ReadArgs(TEMPLATE, opts, NULL)))
  {
    /* Set prefs according to arguments */
    cli_amiga.filename = strdup((char *)opts[OPT_FILE]);

    /* Defaults */
    cli_amiga.frameskip = 0;
    ui_interlace = DEINTERLACE_WEAVEFILTER;
    sound_minfields = 6; /* min fields to try to buffer */
    sound_maxfields = sound_minfields * 2; /* max fields before blocking */

    /* Get use options */
    if (opts[OPT_SIMPLE]) ui_vdpsimple = 1;
    if (opts[OPT_NOSOUND]) { cpuz80_on = 0; sound_fm = 0; sound_psg = 0; }
    if (opts[OPT_NOZ80]) cpuz80_on = 0;
    if (opts[OPT_NOFM]) sound_fm = 0;
    if (opts[OPT_NOPSG]) sound_psg = 0;
    if (opts[OPT_LACE]) {
      char *s = (char *)opts[OPT_LACE];
      if (strcasecmp(s, "bob") == 0) {
        ui_interlace = DEINTERLACE_BOB;
      } else if (strcasecmp(s, "weave") == 0) {
        ui_interlace = DEINTERLACE_WEAVE;
      } else if (strcasecmp(s, "weave-filter") == 0) {
        ui_interlace = DEINTERLACE_WEAVEFILTER;
      }
    }
    if (opts[OPT_LATENCY]) sound_minfields = *((unsigned long *)opts[OPT_LATENCY]);
    if (opts[OPT_FRAGMENTS]) sound_maxfields = *((unsigned long *)opts[OPT_FRAGMENTS]);
    if (opts[OPT_FREQ]) sound_speed = *((unsigned long *)opts[OPT_FREQ]);

    if (opts[OPT_FRAMESKIP]) cli_amiga.frameskip = *((unsigned long *)opts[OPT_FRAMESKIP]);
    if (opts[OPT_JOYPAD]) cli_amiga.joypad = 1;
    if (opts[OPT_REQ]) cli_amiga.req = 1;
    if (opts[OPT_NOSTALGIA]) cli_amiga.nostalgia = 1;

    if (opts[OPT_INTERPRETER]) cli_amiga.interpreter = 1;
    if (opts[OPT_COMPILER]) cli_amiga.compiler = 1;

    if (opts[OPT_SPEEDTEST]) {
      char *s = (char *)opts[OPT_SPEEDTEST];

      if (strncasecmp(s, "disp", 4) == 0) {
        cli_amiga.speedtest = SPEEDTEST_DISP;
      } else if (strncasecmp(s, "null", 4) == 0) {
        cli_amiga.speedtest = SPEEDTEST_NULL;
      } else if (strncasecmp(s, "rend", 4) == 0) {
        cli_amiga.speedtest = SPEEDTEST_REND;
      }

      if (cli_amiga.speedtest) { cpuz80_on = 0; sound_fm = 0; sound_psg = 0; }
    }

    /* Setup generator variables */
    if ((sound_fm == 0) && (sound_psg == 0)) {
      sound_on = 0;
    }

    /* Check the settings to make sure nothing is wrong */
    if ((cli_amiga.frameskip < 0) || (cli_amiga.frameskip > 9)) {
      cli_amiga.frameskip = 0;
    }

    /* If both interpreter and compiler is selected, choose none */
    if ((cli_amiga.interpreter == 1) && (cli_amiga.compiler == 1)) {
      cli_amiga.interpreter = 0;
      cli_amiga.compiler = 0;
    }

    FreeArgs((struct RDArgs *)rdargs);
  } else {
    printf("\n%s\n\nusage: %s <romfile>\n\n", &ver[6], argv[0]);

    printf("SIMPLE/S      : simple VDP emulation.\n");
    printf("FRAMESKIP/N   : frameskip 0-9 (def: 0 = automatic).\n");
    printf("NOSOUND/S     : disable Z80, YM2612 and SN76489 emulation.\n");
    printf("NOZ80/S       : disable Z80 emulation.\n");
    printf("NOFM/S        : disable YM2612 emulation.\n");
    printf("NOPSG/S       : disable SN76489 emulation.\n");
    printf("LACE/K        : select de-interlace mode (bob, weave, weave-filter).\n");
    printf("LATENCY/N     : number of video fields worth of sound to try to keep buffered.\n");
    printf("FRAGMENTS/N   : number of video fields worth of sound to allow buffered.\n");
    printf("FREQ/N        : sound frequency (def: 22050, 11025 for T version).\n");
    printf("JOYPAD/S      : CD32 joypad (+ keyboard for 2nd player) (def: keyboard only).\n");
    printf("REQ/S         : use screenmode requester (def: CGFX BestModeID/PAL Lores).\n\n");
    printf("NOSTALGIA/S   : save the screenshot in Nostalgia:samples/ (def: current directory).\n\n");
    printf("INTERPRETER/S : use optimized interpreter (def: standard interpreter).\n");
    printf("COMPILER/S    : use optimized compiler (def: standard interpreter).\n\n");
    exit(0);
  }
}
