/* Generator is (c) James Ponder, 1997-2001 http://www.squish.net/generator/ */

#include "generator.h"
#include "registers.h"
#include "vdp.h"
#include "cpu68k.h"
#include "mem68k.h"
#include "cpuz80.h"
#include "reg68k.h"
#include "ui.h"
#include "gensound.h"

#include "snprintf.h"

#include "compile.h"
#include <setjmp.h>

/* due to DMA transfers, event_nextevent can be called during an instruction
   cycle (reg68k_external_execute -> instruction -> vdp write -> dma ->
   event_freeze -> event_nextevent).  Be careful */

/* time for next event - update vdp_event - return when to call again */

static inline void event_nextevent(void)
{
  /* call this when it *is* time for the next event as dictated by vdp_event,
     so we switch on it and update vdp_event at the same time */

  switch (vdp_event++) {
  case 0:
  EVENT_NEWLINE:
    LOG_DEBUG1(("%08X due %08X, %d A: %d (cd=%d)",
                cpu68k_clocks, vdp_event_start,
                vdp_line - vdp_visstartline, vdp_reg[10],
                vdp_hskip_countdown));
    if (vdp_line == 0)
      sound_startfield();
    if (vdp_line == (vdp_visstartline - 1)) {
      vdp_vblank = 0;
      vdp_hskip_countdown = vdp_reg[10];
    }
    if ((vdp_nextevent = vdp_event_vint - cpu68k_clocks) > 0)
      break;
    vdp_event++;
  case 1:
    LOG_DEBUG1(("%08X due %08X, %d B: %d (cd=%d)",
                cpu68k_clocks, vdp_event_vint,
                vdp_line - vdp_visstartline, vdp_reg[10],
                vdp_hskip_countdown));
    if (vdp_line == vdp_visendline) {
      vdp_vblank = 1;
      vdp_vsync = 1;
      if (vdp_reg[1] & 1 << 5)
        reg68k_external_autovector(6);  /* vertical interrupt */
    }
    if ((vdp_nextevent = vdp_event_hint - cpu68k_clocks) > 0)
      break;
    vdp_event++;
  case 2:
    LOG_DEBUG1(("%08X due %08X, %d C: %d (cd=%d)",
                cpu68k_clocks, vdp_event_hint,
                vdp_line - vdp_visstartline, vdp_reg[10],
                vdp_hskip_countdown));
    if (vdp_line >= vdp_visstartline && vdp_line < vdp_visendline)
      vdp_hblank = 1;
    if (vdp_line == (vdp_visstartline - 1) || (vdp_line > vdp_visendline)) {
      vdp_hskip_countdown = vdp_reg[10];
      LOG_DEBUG1(("H counter reset to %d", vdp_hskip_countdown));
    }
    if (vdp_reg[0] & 1 << 4) {
      LOG_DEBUG1(("pre = %d", vdp_hskip_countdown));
      if (vdp_hskip_countdown-- == 0) {
        LOG_DEBUG1(("in = %d", vdp_hskip_countdown));
        /* re-initialise counter */
        vdp_hskip_countdown = vdp_reg[10];
        LOG_DEBUG1(("H counter looped to %d", vdp_hskip_countdown));
        if (vdp_line >= vdp_visstartline - 1 && vdp_line < vdp_visendline - 1)
          reg68k_external_autovector(4);        /* horizontal interrupt */
        /* since this game is obviously timing sensitive, we sacrifice
           executing the right number of 68k clocks this frame in order
           to accurately do the moments following H-Int */
        cpu68k_clocks = vdp_event_hint;
      }
      LOG_DEBUG1(("post = %d", vdp_hskip_countdown));
    }
    /* the 68k is frozen for 68k ram to vram copies, see event_freeze */
    if (vdp_dmabytes) {         /* blank mode ? */
      vdp_dmabytes -= (vdp_vblank || !(vdp_reg[1] & 1 << 6))
        ? ((vdp_reg[12] & 1) ? 205 : 167) : ((vdp_reg[12] & 1) ? 18 : 16);
      if (vdp_dmabytes <= 0) {
        vdp_dmabytes = 0;
        vdp_dmabusy = 0;
      }
    }
    if ((vdp_nextevent = vdp_event_hdisplay - cpu68k_clocks) > 0)
      break;
    vdp_event++;
  case 3:
    LOG_DEBUG1(("%08X due %08X, %d D: %d (cd=%d)",
                cpu68k_clocks, vdp_event_hdisplay,
                vdp_line - vdp_visstartline, vdp_reg[10],
                vdp_hskip_countdown));
    ui_line(vdp_line - vdp_visstartline + 1);
    if ((vdp_nextevent = vdp_event_end - cpu68k_clocks) > 0)
      break;
    /* vdp_event++; - not required, we set vdp_event to 0 below */
  case 4:
    /* end of line, do sound, platform stuff */
    LOG_DEBUG1(("%08X due %08X, %d E: %d (cd=%d)",
                cpu68k_clocks, vdp_event_end,
                vdp_line - vdp_visstartline, vdp_reg[10],
                vdp_hskip_countdown));
    if (vdp_line >= vdp_visstartline && vdp_line < vdp_visendline)
      vdp_hblank = 0;
    cpuz80_sync();
    sound_line();
    vdp_line++;
    if (vdp_line == vdp_visendline)
      cpuz80_interrupt();
    if (vdp_line == vdp_totlines) {
      /* the order of these is important */
      sound_endfield(); /* must be before ui_endfield for GYM log and also
                           for avi output */
      ui_endfield();
      vdp_endfield(); /* must be after ui_endfield as it alters state */
      cpuz80_endfield();
      cpu68k_endfield();
      cpu68k_frames++;
    }
    vdp_event_start += vdp_clksperline_68k;
    vdp_event_vint += vdp_clksperline_68k;
    vdp_event_hint += vdp_clksperline_68k;
    vdp_event_hdisplay += vdp_clksperline_68k;
    vdp_event_end += vdp_clksperline_68k;
    vdp_event = 1;
    goto EVENT_NEWLINE;
  }                             /* switch */
}

/*** event_go_interpreter - execute until user decides to quit ***/

t_ipclist *event_interpreter_nextevent(void)
{
  for (;;) {
    if (vdp_nextevent > 0) {
      unsigned int index;
      t_ipclist *list;
      uint32 pc24;
      t_ipc step_ipc;
      t_iib *step_piib;

      if (regs.pending && ((reg68k_sr.sr_int >> 8) & 7) < regs.pending)
        reg68k_internal_autovector(regs.pending);

      pc24 = reg68k_pc & 0xffffff;
      if ((pc24 & 0xff0000) == 0xff0000) {
        /* executing code from RAM, do not use compiled information */
        do {
          step_piib = cpu68k_iibtable[fetchword(reg68k_pc)];
          if (!step_piib)
            ui_err("Invalid instruction (iib assert) @ %08X\n", reg68k_pc);
          cpu68k_ipc(reg68k_pc,
                     mem68k_memptr[(reg68k_pc >> 12) &
                                   0xfff] (reg68k_pc & 0xFFFFFF),
                     step_piib, &step_ipc);
          cpu68k_functable[fetchword(reg68k_pc) * 2 + 1] (&step_ipc);
          cpu68k_clocks += step_piib->clocks;
          vdp_nextevent -= step_piib->clocks;
        }
        while (!step_piib->flags.endblk);
      } else {
        index = (pc24 >> 1) & (LEN_IPCLISTTABLE - 1);
        list = ipclist[index];
        while (list && (list->pc != pc24)) {
          list = list->next;
        }

        if (!list) {
          /* LOG_USER(("Making IPC list @ %08x", pc24)); */
          list = cpu68k_makeipclist(pc24);
          list->next = ipclist[index];
          ipclist[index] = list;
          list->compiled = NULL;
        }

        cpu68k_clocks += list->clocks;
        vdp_nextevent -= list->clocks;

        return list;
      }
    } else {
      regs.pc = reg68k_pc;
      regs.sr = reg68k_sr;

      event_nextevent();

      reg68k_pc = regs.pc;
      reg68k_sr = regs.sr;

      if (gen_quit) {
        return NULL;
      }
    }
  }
}

#if 1

void event_init_interpreter(void)
{
}

void event_go_interpreter(void)
{
  unsigned int index;
  t_ipclist *list;
  t_ipc *ipc;
  uint32 pc24;
  t_ipc step_ipc;
  t_iib *step_piib;
  int clks;
  jmp_buf jb;

  if (!setjmp(jb)) {
    /* move PC and register block into global variables */
    reg68k_pc = regs.pc;
    reg68k_regs = regs.regs;
    reg68k_sr = regs.sr;

    for (;;) {
      while (vdp_nextevent <= 0) {
        event_nextevent();
      }

      if (gen_quit) {
        break;
      }

      clks = vdp_nextevent;

      /* move PC and register block into global variables */
      reg68k_pc = regs.pc;
      reg68k_sr = regs.sr;

      if (regs.pending && ((reg68k_sr.sr_int >> 8) & 7) < regs.pending)
        reg68k_internal_autovector(regs.pending);

      do {
        pc24 = reg68k_pc & 0xffffff;
        if ((pc24 & 0xff0000) == 0xff0000) {
          /* executing code from RAM, do not use compiled information */
          do {
            step_piib = cpu68k_iibtable[fetchword(reg68k_pc)];
            if (!step_piib)
              ui_err("Invalid instruction (iib assert) @ %08X\n", reg68k_pc);
            cpu68k_ipc(reg68k_pc,
                       mem68k_memptr[(reg68k_pc >> 12) &
                                     0xfff] (reg68k_pc & 0xFFFFFF),
                       step_piib, &step_ipc);
            cpu68k_functable[fetchword(reg68k_pc) * 2 + 1] (&step_ipc);
            clks -= step_piib->clocks;
            cpu68k_clocks += step_piib->clocks;
          }
          while (!step_piib->flags.endblk);
        } else {
          index = (pc24 >> 1) & (LEN_IPCLISTTABLE - 1);
          list = ipclist[index];
          while (list && (list->pc != pc24)) {
            list = list->next;
          }
          if (!list) {
            /* LOG_USER(("Making IPC list @ %08x", pc24)); */
            list = cpu68k_makeipclist(pc24);
            list->next = ipclist[index];
            ipclist[index] = list;
          }
          ipc = (t_ipc *) (list + 1);
          do {
            ipc->function(ipc);
            ipc++;
          }
          while (*(int *)ipc);
#if 1
          if (list->norepeat) {
            int num = ((clks + list->clocks - 1) / list->clocks) * list->clocks;
            clks -= num;
            cpu68k_clocks += num;
          } else {
            clks -= list->clocks;
            cpu68k_clocks += list->clocks;
          }
#else
          do {
            clks -= list->clocks;
            cpu68k_clocks += list->clocks;
          } while (list->norepeat && clks > 0);
#endif
        }
      }
      while (clks > 0);

      /* restore global registers back to permanent storage */
      regs.pc = reg68k_pc;
      regs.sr = reg68k_sr;

      vdp_nextevent = clks;
    }

    /* restore global registers back to permanent storage */
    regs.pc = reg68k_pc;
    regs.sr = reg68k_sr;
    longjmp(jb, 1);
  }
}

#else

void event_init_interpreter(void)
{
}

void event_go_interpreter(void)
{
  t_ipclist *list;
  t_ipc *ipc;
  jmp_buf jb;

  if (!setjmp(jb)) {
    /* move PC and register block into global variables */
    reg68k_pc = regs.pc;
    reg68k_regs = regs.regs;
    reg68k_sr = regs.sr;

    for (;;) {
      list = event_interpreter_nextevent();
      if (list == NULL) {
        break;
      }
      ipc = (t_ipc *) (list + 1);
      do {
        ipc->function(ipc);
        ipc++;
      }
      while (*(int *)ipc);
    }

    /* restore global registers back to permanent storage */
    regs.pc = reg68k_pc;
    regs.sr = reg68k_sr;
    longjmp(jb, 1);
  }
}

#endif

/*** event_go_compiler - execute until user decides to quit ***/

static compiler_t *compiler;

void event_init_compiler(void)
{
  compiler = compile_init();
}

void event_go_compiler(void)
{
  jmp_buf jb;

  if (!setjmp(jb)) {
    /* move PC and register block into global variables */
    reg68k_pc = regs.pc;
    reg68k_regs = regs.regs;
    reg68k_sr = regs.sr;

    if (compiler != NULL) {
      compiler->go();
      compile_exit();
    }

    /* restore global registers back to permanent storage */
    regs.pc = reg68k_pc;
    regs.sr = reg68k_sr;
    longjmp(jb, 1);
  }
}

/* called by compiled code */
void *event_compiler_nextevent(void)
{
  t_ipclist *list = event_interpreter_nextevent();

  if (list != NULL) {
    if (list->compiled == NULL) {
      list->compiled = compile_make(list);
    }
    return list->compiled;
  }

  return compiler->exit;
}

void event_compiler_catchup(void)
{
  /* restore global registers back to permanent storage */
  regs.pc = reg68k_pc;
  regs.sr = reg68k_sr;

  while (vdp_nextevent <= 0) {
    event_nextevent();
  }

  if (gen_quit) {
    compile_quit();
  }

  /* move PC and register block into global variables */
  reg68k_pc = regs.pc;
  reg68k_sr = regs.sr;

  if (regs.pending && ((reg68k_sr.sr_int >> 8) & 7) < regs.pending)
    reg68k_internal_autovector(regs.pending);
}

void *event_compiler_compile(void)
{
  unsigned int index;
  t_ipclist *list;
  uint32 pc24;

  pc24 = reg68k_pc & 0xffffff;

  index = (pc24 >> 1) & (LEN_IPCLISTTABLE - 1);
  list = ipclist[index];
  while (list && (list->pc != pc24)) {
    list = list->next;
  }

  if (!list) {
    /* LOG_USER(("Making IPC list @ %08x", pc24)); */
    list = cpu68k_makeipclist(pc24);
    list->next = ipclist[index];
    ipclist[index] = list;
    list->compiled = compile_make(list);
  }

  return list->compiled;
}

void event_compiler_ram(void)
{
  uint32 pc24;
  t_ipc step_ipc;
  t_iib *step_piib;

  pc24 = reg68k_pc & 0xffffff;
  /* executing code from RAM, do not use compiled information */
  do {
    step_piib = cpu68k_iibtable[fetchword(reg68k_pc)];
    if (!step_piib)
      ui_err("Invalid instruction (iib assert) @ %08X\n", reg68k_pc);
    cpu68k_ipc(reg68k_pc,
               mem68k_memptr[(reg68k_pc >> 12) &
                             0xfff] (reg68k_pc & 0xFFFFFF),
               step_piib, &step_ipc);
    cpu68k_functable[fetchword(reg68k_pc) * 2 + 1] (&step_ipc);
    cpu68k_clocks += step_piib->clocks;
    vdp_nextevent -= step_piib->clocks;
  }
  while (!step_piib->flags.endblk);
}
