/*****************************************************************************/
/*     Generator - Sega Genesis emulation - (c) James Ponder 1997-2001       */
/*****************************************************************************/
/*                                                                           */
/* ui_amiga.c - Amiga user interface (c) Mathias "AmiDog" Roslund 2001       */
/*                                                                           */
/*****************************************************************************/

#undef COMPARE /* Define to use compare buffer for gfxmem writes */

#define OLD_TYPEDEFS
typedef float float32;
typedef double float64;

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "generator.h"
#include "cpu68k.h"
#include "cpuz80.h"
#include "ui.h"
#include "vdp.h"
#include "mem68k.h"
#include "gensound.h"
#include "gensoundp.h"
#include "event.h"
#include "ui-console.h"
#include "cli-amiga.h"

#include "timer.h"

#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/Picasso96API.h>
#include <proto/asl.h>
#include <proto/timer.h>

static timer_t *timer;

#define GetSysTimePPC(tv) timer_gettime(timer, tv)
#define SubTimePPC(dst, src) timer_subtime(timer, dst, src);

uint8 ui_vdpsimple = 0;         /* 0=raster, 1=cell based plotter */
static uint8 ui_plotfield = 0;  /* flag indicating plotting this field */
t_interlace ui_interlace = DEINTERLACE_WEAVEFILTER;

struct Screen *screen;
struct Window *window;
static UWORD *EmptyPointer;
static struct BitMap agabitmap;

static int aga = 1;
static ULONG rgbtab[1+3*256+1];
static unsigned char *linedata, *pal, *framedata;
static unsigned char *cmp_buf;

/* CGFX variables */
static unsigned long pixfmt;
static unsigned char *baseaddress;
static int bpr;

#define SCREEN_W (320)
#define SCREEN_H (240)

void close_libraries(void)
{
}

int open_libraries(void)
{
  return 0;
}

void close_screen(void)
{
  if (window != NULL) {
    CloseWindow(window);
    window = NULL;
  }
  if (screen != NULL) {
    CloseScreen(screen);
    screen = NULL;
  }

  if (aga) {
    FreeRaster(agabitmap.Planes[0], SCREEN_W, SCREEN_H*8);
  }

  soundp_stop();
}

int open_screen(void)
{
  struct ScreenModeRequester *screenreq;
  struct DimensionInfo dim;
  unsigned long dispid;
  int i;

  if (cli_amiga.req) {
    int width = 0, height = 0, depth = 0;
    char asl_title[32];

    sprintf(asl_title, "%dx%d", SCREEN_W, SCREEN_H);
    screenreq = AllocAslRequestTags(ASL_ScreenModeRequest, TAG_END);
    if (screenreq != NULL) {
      if (AslRequestTags(screenreq, ASLSM_TitleText, (ULONG)asl_title, TAG_END)) {
        dispid = screenreq->sm_DisplayID;

        width = p96GetModeIDAttr(dispid, P96IDA_WIDTH);
        height = p96GetModeIDAttr(dispid, P96IDA_HEIGHT);
        depth = p96GetModeIDAttr(dispid, P96IDA_DEPTH);
        aga = 0;

        if (aga == 1) {
          if (GetDisplayInfoData(NULL, &dim, sizeof(dim), DTAG_DIMS, dispid) == 0) {
            dispid = 0;
          } else {
            width = dim.Nominal.MaxX + 1;
            height = dim.Nominal.MaxY + 1;
            depth = 8;
          }
        }
      } else {
        dispid = 0;
      }
      FreeAslRequest(screenreq);
    } else {
      dispid = 0;
    }
    if ((dispid == 0) || (width != 320) || (depth != 8)) {
      printf("You must select an 8bit screenmode with a width of 320 pixels!\n");
      return -1;
    }
  } else {
    dispid = p96BestModeIDTags(P96BIDTAG_NominalWidth, SCREEN_W,
                             P96BIDTAG_NominalHeight, SCREEN_H,
                             P96BIDTAG_Depth, 8,
                             TAG_DONE);

    if (dispid == INVALID_ID) {
      printf("p96BestModeID failed\n");
      return -1;
    }
    aga = 0;
  }

  if (aga) {
    /* Allocate resources */
    unsigned char *temp;
    temp = (unsigned char *)AllocRaster(SCREEN_W, SCREEN_H*8);
    memset(temp, 0, SCREEN_W*SCREEN_H);

    InitBitMap(&agabitmap, 8, SCREEN_W, SCREEN_H); 

    for(i=0; i<8; i++) {
      agabitmap.Planes[i] = temp + i * RASSIZE(SCREEN_W, SCREEN_H);
    }

    /* Open screen */
    screen = OpenScreenTags(NULL,
             SA_BitMap, (int)&agabitmap,
             SA_Width, SCREEN_W,
             SA_Height, SCREEN_H,
             SA_Depth, 8,
             SA_Quiet, TRUE,
             SA_ShowTitle, FALSE,
             SA_Type, CUSTOMSCREEN,
             SA_DisplayID, dispid,
             TAG_DONE);
  } else {
    /* Open screen */
    screen = OpenScreenTags(NULL,
             SA_Width, SCREEN_W,
             SA_Height, SCREEN_H,
             SA_Depth, 8,
             SA_Quiet, TRUE,
             SA_ShowTitle, FALSE,
             SA_Type, CUSTOMSCREEN,
             SA_DisplayID, dispid,
             TAG_DONE);
  }

  /* Could the screen be opened? */
  if (screen == NULL) {
    printf("OpenScreen failed\n");
    if (aga) {
      FreeRaster(agabitmap.Planes[0], SCREEN_W, SCREEN_H*8);
    }
    return -1;
  }

  if (aga) {
    baseaddress = malloc(SCREEN_W * SCREEN_H);
    memset(baseaddress, 0, (SCREEN_W * SCREEN_H));
    bpr = 320;
  } else {
    pixfmt = p96GetBitMapAttr(screen->RastPort.BitMap, P96BMA_RGBFORMAT);
    baseaddress = (unsigned char *)p96GetBitMapAttr(screen->RastPort.BitMap, P96BMA_MEMORY);
    bpr = p96GetBitMapAttr(screen->RastPort.BitMap, P96BMA_BYTESPERROW);
  }

  /* Open window */
  window = OpenWindowTags(NULL,
           WA_CustomScreen, (int)screen,
           WA_Width, SCREEN_W,
           WA_Height, SCREEN_H,
           WA_IDCMP, IDCMP_RAWKEY,
           WA_Backdrop, TRUE,
           WA_Borderless, TRUE,
           WA_Activate, TRUE,
           TAG_DONE);

  /* Could the window be opened? */
  if (window == NULL) {
    CloseScreen(screen);
    if (aga) {
      FreeRaster(agabitmap.Planes[0], SCREEN_W, SCREEN_H*8);
    }
    printf("OpenWindow failed\n");
    return -1;
  }

  /* Set a 3:3:2 palette */
  for (i=0; i<256; i++) {
    rgbtab[1+i*3+0] = ((i & 0xe0))<<24;
    rgbtab[1+i*3+1] = ((i & 0x1c) << 3)<<24;
    rgbtab[1+i*3+2] = ((0xe0*((i & 0x03) << 6))/0xc0)<<24;
  }

  rgbtab[0]=(256<<16)|0;
  rgbtab[1+256*3]=0;
  LoadRGB32(&screen->ViewPort, rgbtab);

  /* Hide the pointer */
  EmptyPointer = AllocVec(512, MEMF_CHIP|MEMF_CLEAR);
  SetPointer(window, EmptyPointer, 1, 1, 0, 0);

  return 0;
}

char *ui_setinfo(t_cartinfo *cartinfo)
{
  char buffer[128];

  printf("Console:\t%s\n", cartinfo->console);
  printf("Copyright:\t%s\n", cartinfo->copyright);
  printf("Name domestic:\t%s\n", cartinfo->name_domestic);
  printf("Name overseas:\t%s\n", cartinfo->name_overseas);
  printf("Version:\t%s\n", cartinfo->version);
  printf("Memo:\t%s\n", cartinfo->memo);
  sprintf(buffer, "%08x", cartinfo->checksum);
  printf("Checksum:\t%s\n", buffer);
  printf("Prodtype:\t");
  if (cartinfo->prodtype == pt_game) {
    printf("Game\n");
  } else if (cartinfo->prodtype == pt_education) {
    printf("Education");
  } else {
    printf("Unknown");
  }
  return NULL;
}

void ui_simpleplot(void)
{
  unsigned long *src, *dst, *cmp;
  unsigned short color;
  unsigned char *ptr, *framestart;
  int i, line;

  /* simple mode - entire frame done here */
  framestart = framedata + (320 * 8) + 8;
  vdp_renderframe(framestart, 320 + 16); /* plot frame */

  if (cli_amiga.speedtest == SPEEDTEST_REND) {
    return;
  }

  ptr = pal;
  /* Loop through the colors */
  for (i=0; i<64; i++) {
    /* Has the color changed? */
    if (vdp_cramf[i]) {
      color = ((unsigned short *)vdp_cram)[i];

      /* xxxxbbbxgggxrrrx -> rrrgggbb */
      color = ((color & 0x000e) << 4) | ((color & 0x00e0) >> 3) | ((color & 0x0c00) >> 10);

      /* Color 0-63 */
      ptr[0] = color;

      /* Color 64-127 */
      ptr[64] = ((color >> 1) | 0x92);

      /* Color 128-191 */
      ptr[128] = ((color >> 1) & 0x6d);

      /* Set the color to "not changed" */
      vdp_cramf[i] = 0;
    }
    ptr++;
  }

  cmp = (unsigned long *)cmp_buf;
  src = (unsigned long *)framestart;
  dst = (unsigned long *)baseaddress;

  for (line=0; line<224; line++) {
    for (i=0; i<80; i++) {
      unsigned long col, val = *src++;

#if 1 /* FIXME: Why is this correct on my Radeon 7000 ??? */
      col = (pal[((val >> 24) & 0xff)] << 16) |
            (pal[((val >> 16) & 0xff)] << 24) |
            (pal[((val >>  8) & 0xff)] <<  0) |
            (pal[((val >>  0) & 0xff)] <<  8) ;
#else
      col = (pal[((val >> 24) & 0xff)] << 24) |
            (pal[((val >> 16) & 0xff)] << 16) |
            (pal[((val >>  8) & 0xff)] <<  8) |
            (pal[((val >>  0) & 0xff)] <<  0) ;
#endif

#ifndef COMPARE
      *dst++ = col;
#else
      if (*cmp != col) {
        *dst = col;
        *cmp = col;
      }
      dst++;
      cmp++;
#endif
    }
    src += 4; /* 16 bytes = 4 longs */
    dst += (bpr - 320) >> 2;
  }
}

void ui_line(int line)
{
  if (!ui_plotfield)
    return;

  if (line < 0 || line >= (int)vdp_vislines)
    return;

  if (ui_vdpsimple) {
    if (line == (int)(vdp_vislines >> 1)) 
      /* if we're in simple cell-based mode, plot when half way
       * down screen */
      ui_simpleplot();
    return;
  } else {
    unsigned long *src, *dst, *cmp;
    unsigned short color;
    unsigned char *ptr;
    int i;

    /* Render one line of graphics */
    if (((vdp_reg[12]>>1) & 3) == 3)
      vdp_renderline(line, (uint8 *)linedata, vdp_oddframe);
    else
      vdp_renderline(line, (uint8 *)linedata, 0);

    if (cli_amiga.speedtest == SPEEDTEST_REND) {
      return;
    }

    ptr = pal;
    /* Loop through the colors */
    for (i=0; i<64; i++) {
      /* Has the color changed? */
      if (vdp_cramf[i]) {
        color = ((unsigned short *)vdp_cram)[i];

        /* xxxxbbbxgggxrrrx -> rrrgggbb */
        color = ((color & 0x000e) << 4) | ((color & 0x00e0) >> 3) | ((color & 0x0c00) >> 10);

        /* Color 0-63 */
        ptr[0] = color;

        /* Color 64-127 */
        ptr[64] = ((color >> 1) | 0x92);

        /* Color 128-191 */
        ptr[128] = ((color >> 1) & 0x6d);

        /* Set the color to "not changed" */
        vdp_cramf[i] = 0;
      }
      ptr++;
    }

    cmp = (unsigned long *)(cmp_buf + (320 * line));
    src = (unsigned long *)linedata;
    dst = (unsigned long *)(baseaddress + (bpr * line));

    for (i=0; i<80; i++) {
      unsigned long col, val = *src++;

#if 1 /* FIXME: Why is this correct on my Radeon 7000 ??? */
      col = (pal[((val >> 24) & 0xff)] << 16) |
            (pal[((val >> 16) & 0xff)] << 24) |
            (pal[((val >>  8) & 0xff)] <<  0) |
            (pal[((val >>  0) & 0xff)] <<  8) ;
#else
      col = (pal[((val >> 24) & 0xff)] << 24) |
            (pal[((val >> 16) & 0xff)] << 16) |
            (pal[((val >>  8) & 0xff)] <<  8) |
            (pal[((val >>  0) & 0xff)] <<  0) ;
#endif

#ifndef COMPARE
      *dst++ = col;
#else
      if (*cmp != col) {
        *dst = col;
        *cmp = col;
      }
      dst++;
      cmp++;
#endif
    }
  }
}

void ui_final(void)
{
  close_screen();
  close_libraries();
}

/*** ui_endfield - end of field reached ***/

#define TWOMERGES(a, b, c, d, shift, mask) \
     { \
     temp1 = b; \
     temp2 = d; \
     temp1 >>= shift; \
     temp2 >>= shift; \
     temp1 ^= a; \
     temp2 ^= c; \
     temp1 &= mask; \
     temp2 &= mask; \
     a ^= temp1; \
     c ^= temp2; \
     temp1 <<= shift; \
     temp2 <<= shift; \
     b ^= temp1; \
     d ^= temp2; \
     }

void c2p8(void *chunky, void *bitplanes, int chunkyx, int chunkyy, int xoffset, int yoffset, int bitplanesize)
{
     ULONG     *c;
     ULONG     temp1, temp2;
     ULONG     *p0, *p1, *p2, *p3, *p4, *p5, *p6, *p7;
     ULONG     d0, d1, d2, d3, d4, d5, d6, d7;
     ULONG     t0, t1, t2, t3, t4, t5, t6, t7;
     ULONG     xadd, addx;
     ULONG     i, j=0;

     c = (ULONG *) chunky;
     p0 = (ULONG *) (&(((char *) bitplanes)[(yoffset*chunkyx + yoffset*xoffset*2 + xoffset) >> 3]));
     p1 = (ULONG *) (&(((char *) p0)[bitplanesize]));
     p2 = (ULONG *) (&(((char *) p1)[bitplanesize]));
     p3 = (ULONG *) (&(((char *) p2)[bitplanesize]));
     p4 = (ULONG *) (&(((char *) p3)[bitplanesize]));
     p5 = (ULONG *) (&(((char *) p4)[bitplanesize]));
     p6 = (ULONG *) (&(((char *) p5)[bitplanesize]));
     p7 = (ULONG *) (&(((char *) p6)[bitplanesize]));

     addx=chunkyx / (sizeof (ULONG) *8);
     xadd=xoffset * 2 / (sizeof (ULONG) * 8);
     i = chunkyx * chunkyy / (sizeof (ULONG) * 8);

     d0 = *c++;
     d1 = *c++;
     d2 = *c++;
     d3 = *c++;
     d4 = *c++;
     d5 = *c++;
     d6 = *c++;
     d7 = *c++;

     TWOMERGES(d0, d4, d1, d5, 16, 0x0000ffff);
     TWOMERGES(d2, d6, d3, d7, 16, 0x0000ffff);

     TWOMERGES(d0, d2, d1, d3, 8, 0x00ff00ff);
     TWOMERGES(d4, d6, d5, d7, 8, 0x00ff00ff);

     TWOMERGES(d0, d1, d2, d3, 4, 0x0f0f0f0f);
     TWOMERGES(d4, d5, d6, d7, 4, 0x0f0f0f0f);

     TWOMERGES(d0, d4, d1, d5, 2, 0x33333333);
     TWOMERGES(d2, d6, d3, d7, 2, 0x33333333);

     TWOMERGES(d0, d2, d1, d3, 1, 0x55555555);
     TWOMERGES(d4, d6, d5, d7, 1, 0x55555555);

     t0 = d7;
     t1 = d5;
     t2 = d3;
     t3 = d1;
     t4 = d6;
     t5 = d4;
     t6 = d2;
     t7 = d0;

     while (--i)
     {
          d0 = *c++;
          d1 = *c++;
          d2 = *c++;
          d3 = *c++;
          d4 = *c++;
          d5 = *c++;
          d6 = *c++;
          d7 = *c++;

          *p0++ = t0;

          TWOMERGES(d0, d4, d1, d5, 16, 0x0000ffff);
          TWOMERGES(d2, d6, d3, d7, 16, 0x0000ffff);

          *p1++ = t1;

          TWOMERGES(d0, d2, d1, d3, 8, 0x00ff00ff);
          TWOMERGES(d4, d6, d5, d7, 8, 0x00ff00ff);

          *p2++ = t2;

          TWOMERGES(d0, d1, d2, d3, 4, 0x0f0f0f0f);
          TWOMERGES(d4, d5, d6, d7, 4, 0x0f0f0f0f);

          *p3++ = t3;

          TWOMERGES(d0, d4, d1, d5, 2, 0x33333333);
          TWOMERGES(d2, d6, d3, d7, 2, 0x33333333);

          *p4++ = t4;

          TWOMERGES(d0, d2, d1, d3, 1, 0x55555555);
          TWOMERGES(d4, d6, d5, d7, 1, 0x55555555);

          *p5++ = t5;
          *p6++ = t6;
          *p7++ = t7;

          t0 = d7;
          t1 = d5;
          t2 = d3;
          t3 = d1;
          t4 = d6;
          t5 = d4;
          t6 = d2;
          t7 = d0;

          j++;

          if (j==addx)
          {
            p0+=xadd;
            p1+=xadd;
            p2+=xadd;
            p3+=xadd;
            p4+=xadd;
            p5+=xadd;
            p6+=xadd;
            p7+=xadd;
            j=0;
          }
     }

     *p0++ = t0;
     *p1++ = t1;
     *p2++ = t2;
     *p3++ = t3;
     *p4++ = t4;
     *p5++ = t5;
     *p6++ = t6;
     *p7++ = t7;
}

void check_idcmp(void);

#define SPEEDTEST_FRAMES 5000
static struct timeval st, dt;

void ui_endfield(void)
{
  if (cli_amiga.speedtest > 0) {
    if (cpu68k_frames >= SPEEDTEST_FRAMES) {
      double time;

      GetSysTimePPC(&dt);
      SubTimePPC(&dt, &st);

      time = (double)dt.tv_secs + ((double)dt.tv_micro / 1000000.0);
      printf("Time: %.3f, Frames: %d, FPS: %.3f\n", time, SPEEDTEST_FRAMES, (double)SPEEDTEST_FRAMES / time);

      gen_quit = 1; /* QUIT */
    }

    if (cli_amiga.speedtest == SPEEDTEST_NULL) {
      ui_plotfield = 0;
    } else {
      ui_plotfield = 1;
    }

    return; /* FIXME: with this I can not check AGA performance */
  }

  if (ui_plotfield) {
    if (aga) {
      c2p8((void *)baseaddress, (void *)agabitmap.Planes[0], 320, 224, 0, 0, RASSIZE(320, 240));
    }

    if ((cli_amiga.speedtest != SPEEDTEST_NONE) ||
       ((cpuz80_on == 0) && (sound_fm == 0) && (sound_psg == 0))) {
      static struct timeval tt, ct; /* target time and current time */
      int us;

      /* no audio taking care of sync, do it here */

      if (cpu68k_frames == 0) {
        timer_gettime(timer, &tt);
      } else {
        int sleep;

        timer_gettime(timer, &ct);
        sleep = ((int)tt.tv_secs - (int)ct.tv_secs) * 1000000 + ((int)tt.tv_micro - (int)ct.tv_micro);
//        printf("sleep %d\n", sleep);

        if (sleep > 0) {
          timer_usleep(timer, sleep);
        }
      }

      us = tt.tv_micro + ((1000000 * (cli_amiga.frameskip > 0 ? cli_amiga.frameskip : 1)) / vdp_framerate);
      tt.tv_micro = (us % 1000000);
      tt.tv_secs += (us / 1000000);
    }
  }

  if ((cpu68k_frames % (cli_amiga.frameskip > 0 ? cli_amiga.frameskip : 1)) == 0) {
    ui_plotfield = 1;
  } else {
    ui_plotfield = 0;
  }

  check_idcmp();
}

void check_pad(void)
{
}

/* PCX screenshot saver */
void write_pcx(void)
{
  unsigned char buffer[320*240+256*3+128]; /* X*Y + Palette + Header */
  unsigned char *dst = buffer + 128;
  int start, stop, x, y, i;
  char *name, *file;
  FILE *fp;

  /* Get filename */
  stop = strlen(cli_amiga.filename) - 1;

  while((stop > 0) && (cli_amiga.filename[stop] != '.'))
    stop--;

  if (stop == 0)
    return;

  start = stop;

  while((start > 0) && (cli_amiga.filename[start] != '/') && (cli_amiga.filename[start] != ':'))
    start--;

  if (start == 0) {
    name = (char *)malloc(stop + 5);
    memset(name, 0, (stop + 5));
    strncpy(name, cli_amiga.filename, (stop + 1));
  }
  else {
    name = (char *)malloc(stop - start + 4);
    memset(name, 0, (stop - start + 4));
    for (i = 0; i < (stop - start); i++)
      name[i] = cli_amiga.filename[start + i + 1];
  }

  if (cli_amiga.nostalgia) {
    strcat(name, "smp");
    file = (char *)malloc(strlen(name) + 19);
    memset(file, 0, (strlen(name) + 19));
    strcpy(file, "Nostalgia:samples/");
    strcat(file, name);
  } else {
    strcat(name, "pcx");
    file = strdup(name);
  }

  /* Header (little endian) */
  memset(buffer, 0, sizeof(buffer));

  buffer[ 0] = 10;  /* PCX file */
  buffer[ 1] = 5;   /* 8bit */
  buffer[ 2] = 1;   /* RLE */
  buffer[ 3] = 8;   /* 8bit */

  buffer[ 9] = 1;   /* Window max width */
  buffer[ 8] = 63;  /* Window max width */
  buffer[11] = 0;   /* Window max height */
  buffer[10] = 239; /* Window max height */

  buffer[13] = 1;   /* Width */
  buffer[12] = 64;  /* Width */
  buffer[15] = 0;   /* Height */
  buffer[14] = 240; /* Height */

  buffer[65] = 1;          /* Colorplanes */
  buffer[67] = buffer[13]; /* Bytes per line */
  buffer[66] = buffer[12]; /* Bytes per line */
  buffer[69] = 0;          /* Color */
  buffer[68] = 1;          /* Color */

  for (y = 0; y < 240; y++) {
    unsigned char *src, last, this;
    int total = 0, count = 1;

    /* Init */
    src = cmp_buf + (y * 320); /* WARNING */
    last = *(src);

    /* Encode one line */
    for (x = 1; x < 320; x++) {
      this = *(++src);
      if (this == last) {
        count++;
        if (count == 63) {
          *dst++ = (count | 0xc0);
          *dst++ = last;
          total += 2;
          count = 0;
        }
      } else {
        if (count) {
          if ((count == 1) && ((last & 0xc0) != 0xc0)) {
            *dst++ = last;
            total++;
          } else {
            *dst++ = (count | 0xc0);
            *dst++ = last;
            total += 2;
          }
        }
        last = this;
        count = 1;
      }
    }

    if (count) {
      if ((count == 1) && ((last & 0xc0) != 0xc0)) {
        *dst++ = last;
      } else {
        *dst++ = (count | 0xc0);
        *dst++ = last;
      }
    }
  }

  *dst++ = 12; /* 8bit palette is to follow */

  for (i=0; i<256; i++) {
    *dst++ = rgbtab[1+i*3+0] >> 24;
    *dst++ = rgbtab[1+i*3+1] >> 24;
    *dst++ = rgbtab[1+i*3+2] >> 24;
  }

  /* Save the file */
  fp = fopen(file, "w");

  if (fp != NULL) {
    fwrite(buffer, 1, (int)(dst-buffer), fp);
    fclose(fp);
  }
}

void check_idcmp(void)
{
  struct IntuiMessage *imsg = NULL;
  ULONG imCode, imClass;
  int type = 0, pad = 0;

  /* Check for IDCMP messages */
  if((imsg=(struct IntuiMessage *)GetMsg(window->UserPort))) {
    imClass = imsg->Class;
    imCode = imsg->Code;
    ReplyMsg((struct Message *)imsg);

    /* KeyPress and KeyReleases */
    if(imClass==IDCMP_RAWKEY) {
      if (imCode == 0xdf) /* HELP */
        write_pcx();

      if(imCode<128) /* Key is being pressed */
        type = 1;

      imCode &= 0x7F;

      switch(imCode) {
        case 69: /* ESC, Quit */
          gen_quit = 1;
        break;

        case 32: /* A */
          mem68k_cont[pad].a = type;
        break;

        case 53: /* B */
        case 33: /* S */
          mem68k_cont[pad].b = type;
        break;

        case 55: /* C */
        case 34: /* D */
          mem68k_cont[pad].c = type;
        break;

        case 79: /* Arrow LEFT */
          mem68k_cont[pad].left = type;
        break;

        case 76: /* Arrow UP */
          mem68k_cont[pad].up = type;
        break;

        case 78: /* Arrow RIGHT */
          mem68k_cont[pad].right = type;
        break;

        case 77: /* Arrow DOWN */
          mem68k_cont[pad].down = type;
        break;

        case 68: /* Enter */
          mem68k_cont[pad].start = type;
        break;

        case 1: case 2: case 3:
        case 4: case 5: case 6:
        case 7: case 8: case 9: /* 1 - 9 */
          cli_amiga.frameskip = imCode;
        break;

        case 10: /* 0 */
          cli_amiga.frameskip = 0;
        break;

        case 84: /* F5 */
          ui_vdpsimple = 0;
        break;

        case 85: /* F6 */
          ui_vdpsimple = 1;
        break;

        default:
        break;
      }
    }
  }
}

extern void event_init_interpreter(void);
extern void event_go_interpreter(void);
extern void event_init_compiler(void);
extern void event_go_compiler(void);

int ui_loop(void)
{
  char *retstr;

  if (open_screen() < 0) {
    close_libraries();
    return 0;
  }

  if ((retstr = gen_loadimage(cli_amiga.filename)))
    fprintf(stderr, retstr);

  if (cli_amiga.speedtest == SPEEDTEST_NULL) {
    ui_plotfield = 0;
  } else {
    ui_plotfield = 1;
  }

  gen_quit = 0;
  if (cli_amiga.interpreter == 1) {
printf("OPTIMIZED INTERPRETER\n");
    event_init_interpreter();
    GetSysTimePPC(&st);
    event_go_interpreter();
  } else if (cli_amiga.compiler == 1) {
printf("OPTIMIZED COMPILER\n");
    event_init_compiler();
    GetSysTimePPC(&st);
    event_go_compiler();
  } else {
printf("STANDARD INTERPRETER\n");
    GetSysTimePPC(&st);
    while (gen_quit == 0) {
      event_doframe();
    }
  }
  gen_quit = 0;

  ui_final();

  timer_exit(timer);

  return 0;
}

int ui_init(int argc, char *argv[])
{
  if (open_libraries() < 0) {
    close_libraries();
    return 1;
  }

  linedata = malloc(320);
  pal = malloc(256); /* Not really required, 192 should be enough */
  framedata = malloc(((320 + 16) * (240 + 16)));
  cmp_buf = malloc(320 * 240);

  memset(linedata, 0, 320);
  memset(pal, 0, 256);
  memset(framedata, 0, ((320 + 16) * (240 + 16)));
  memset(cmp_buf, 0, (320 * 240));

  /* Deal with commandline arguments */
  sound_speed = 11025;
  handle_arguments(argc, argv);

  timer = timer_init();

  return 0;
}

/*** ui_err - log error message and quit ***/

void ui_err(const char *text, ...)
{
  va_list ap;

  printf("ERROR: ");

  va_start(ap, text);
  vfprintf(stderr, text, ap);
  va_end(ap);
  putc(10, stderr);
  ui_final();
}

/*** AmigaOS version doesn't implement music log yet ***/

void ui_musiclog(uint8 *data, unsigned int length)
{
  (void)data;
  (void)length;
}
