/* Generator is (c) James Ponder, 1997-2001 http://www.squish.net/generator/ */

/* user interface for AmigaOS (15 bit) */

/* this part of generator is 'slaved' from the ui-console.c file - only
   platform specific code is contained here */

#include <stdio.h>
#include <stdlib.h>

#include "generator.h"

#include "uip.h"
#include "ui.h"
#include "ui-console.h"
#include "gensound.h"
#include "gensoundp.h"
#include "vdp.h"
#include "cpu68k.h"
#include "mem68k.h"
#include "cli-amiga.h"

#include <exec/types.h>
#include <exec/nodes.h>
#include <exec/lists.h>
#include <exec/memory.h>

#include <powerup/ppcproto/intuition.h>
#include <powerup/ppclib/interface.h>
#include <powerup/ppclib/time.h>
#include <powerup/gcclib/powerup_protos.h>
#include <powerup/ppcproto/exec.h>
#include <powerup/ppcinline/graphics.h>
#include <powerup/ppcproto/dos.h>
#include <powerup/ppcproto/asl.h>
#include <powerup/ppcproto/gadtools.h>
#include <powerup/ppcproto/lowlevel.h>
#include <cybergraphx/cybergraphics.h>
#include <powerup/ppcinline/cybergraphics.h>

/*** forward reference declarations ***/

/*** static variables ***/

static uint8 uip_displaybanknum = 0;    /* view this one, write to other one */
static t_uipinfo *uip_uipinfo = NULL;   /* uipinfo */
static unsigned char *uip_bank0, *uip_bank1, *uip_writebank;
static int depth = 15;

struct IntuitionBase *IntuitionBase = NULL;
struct GfxBase *GfxBase = NULL;
struct Library *AslBase = NULL;
struct Library *CyberGfxBase = NULL;
struct Library *LowLevelBase = NULL;
struct Screen *screen;
struct Window *window;
UWORD *EmptyPointer;
struct ScreenModeRequester *screenreq;

/* CGFX variables */
unsigned long pixfmt;
unsigned char *baseaddress;
int bpr;

void close_libraries()
{
  if (LowLevelBase != NULL)
    CloseLibrary(LowLevelBase);
  if (CyberGfxBase != NULL)
    CloseLibrary(CyberGfxBase);
  if (AslBase != NULL)
    CloseLibrary(AslBase);
  if (GfxBase != NULL)
    CloseLibrary(GfxBase);
  if (IntuitionBase != NULL)
    CloseLibrary(IntuitionBase);
}

int open_libraries()
{
  IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library", 39);
  if (IntuitionBase == NULL) {
    printf("Could not open intuition.library v39 or later!\n");
    return -1;
  }

  GfxBase = (struct GfxBase *)OpenLibrary("graphics.library", 39);
  if (GfxBase == NULL) {
    printf("Could not open graphics.library v39 or later!\n");
    return -1;
  }

  AslBase = OpenLibrary("asl.library", 39);
  if (AslBase == NULL) {
    printf("Could not open asl.library v39 or later!\n");
    return -1;
  }

  CyberGfxBase = OpenLibrary("cybergraphics.library", 40);
  if (CyberGfxBase == NULL) {
    printf("Could not open cybergraphics.library v40 or later!\n");
    return -1;
  }

  LowLevelBase = OpenLibrary("lowlevel.library", 39);

  return 0;
}

void close_screen()
{
  if (window != NULL) {
    CloseWindow(window);
    window = NULL;
  }
  if (screen != NULL) {
    CloseScreen(screen);
    screen = NULL;
  }

  soundp_stop();

  close_libraries();
}

#define SCREEN_W (640)
#define SCREEN_H (480 * 2)

int open_screen()
{
  struct ScreenModeRequester *screenreq;
  int width = 0, height = 0;
  unsigned long dispid = 0;

  if (cli_amiga.req) {
    char asl_title[32];
    sprintf(asl_title, "%dx%d", SCREEN_W, SCREEN_W);
    screenreq = AllocAslRequestTags(ASL_ScreenModeRequest, TAG_END);
    if (screenreq != NULL) {
      if (AslRequestTags(screenreq, ASLSM_TitleText, asl_title, TAG_END)) {
        dispid = screenreq->sm_DisplayID;

        if (IsCyberModeID(dispid)) {
          width = GetCyberIDAttr(CYBRIDATTR_WIDTH, dispid);
          height = GetCyberIDAttr(CYBRIDATTR_HEIGHT, dispid);
          depth = GetCyberIDAttr(CYBRIDATTR_DEPTH, dispid);
        }
      } else {
        dispid = 0;
      }
      FreeAslRequest(screenreq);
    } else {
      dispid = 0;
    }
    if ((dispid == 0) || (width != 640) || (height != 480) || ((depth != 15) && (depth != 16))) {
      printf("You must select an 15/16bit 640x480 screenmode!\n");
      return -1;
    }
  } else {
    depth = 15;
    dispid = BestCModeIDTags(CYBRBIDTG_NominalWidth, 640,
                             CYBRBIDTG_NominalHeight, 480,
                             CYBRBIDTG_Depth, depth,
                             TAG_DONE);

    if (dispid == INVALID_ID) {
      depth = 16;
      dispid = BestCModeIDTags(CYBRBIDTG_NominalWidth, 640,
                               CYBRBIDTG_NominalHeight, 480,
                               CYBRBIDTG_Depth, depth,
                               TAG_DONE);
    }

    if (dispid == INVALID_ID) {
      printf("BestCModeID failed\n");
      return -1;
    }
  }

  /* Open screen */
  screen = OpenScreenTags(NULL,
           SA_Width, SCREEN_W,
           SA_Height, SCREEN_H,
           SA_Depth, depth,
           SA_Quiet, TRUE,
           SA_ShowTitle, FALSE,
           SA_Type, CUSTOMSCREEN,
           SA_DisplayID, dispid,
           TAG_DONE);

  /* Could the screen be opened? */
  if(screen == NULL) {
    printf("OpenScreen failed\n");
    return -1;
  }

  /* Get CGFX screenmode information */
  UnLockBitMap(LockBitMapTags(screen->RastPort.BitMap,
                              LBMI_PIXFMT, (ULONG)&pixfmt,
                              LBMI_BASEADDRESS, (ULONG)&baseaddress,
                              LBMI_BYTESPERROW, (ULONG)&bpr,
                              TAG_END));

  /* Open window */
  window = OpenWindowTags(NULL,
           WA_CustomScreen, (int)screen,
           WA_Width, SCREEN_W,
           WA_Height, SCREEN_H,
           WA_IDCMP, IDCMP_RAWKEY,
           WA_Backdrop, TRUE,
           WA_Borderless, TRUE,
           WA_Activate, TRUE,
           TAG_DONE);

  /* Could the window be opened? */
  if(window == NULL) {
    CloseScreen(screen);
    printf("OpenWindow failed\n");
    return -1;
  }

  FillPixelArray(&screen->RastPort, 0, 0, SCREEN_W, SCREEN_H, 0L);

  /* Hide the pointer */
  EmptyPointer = AllocVec(512, MEMF_CHIP|MEMF_CLEAR);
  SetPointer(window, EmptyPointer, 1, 1, 0, 0);

  return 0;
}

#define KEY_ESC 69

#define KEY_F1 80
#define KEY_F2 81
#define KEY_F3 82
#define KEY_F4 83
#define KEY_F5 84
#define KEY_F6 85
#define KEY_F7 86
#define KEY_F8 87
#define KEY_F9 88
#define KEY_F10 89

#define KEY_1 1
#define KEY_2 2
#define KEY_3 3
#define KEY_4 4
#define KEY_5 5
#define KEY_6 6
#define KEY_7 7
#define KEY_8 8
#define KEY_9 9
#define KEY_0 10

#define KEY_Q 16
#define KEY_W 17
#define KEY_E 18
#define KEY_R 19
#define KEY_T 20
#define KEY_Y 21
#define KEY_U 22
#define KEY_I 23
#define KEY_O 24
#define KEY_P 25

#define KEY_A 32
#define KEY_S 33
#define KEY_D 34
#define KEY_F 35
#define KEY_G 36
#define KEY_H 37
#define KEY_J 38
#define KEY_K 39
#define KEY_L 40

#define KEY_Z 49
#define KEY_X 50
#define KEY_C 51
#define KEY_V 52
#define KEY_B 53
#define KEY_N 54
#define KEY_M 55

#define KEY_0_PAD 15
#define KEY_1_PAD 29
#define KEY_2_PAD 30
#define KEY_3_PAD 31
#define KEY_4_PAD 45
#define KEY_5_PAD 46
#define KEY_6_PAD 47
#define KEY_7_PAD 61
#define KEY_8_PAD 62
#define KEY_9_PAD 63

#define KEY_LEFT 79
#define KEY_RIGHT 78
#define KEY_UP 76
#define KEY_DOWN 77

#define KEY_SLASH_PAD 92
#define KEY_ASTERISK_PAD 93
#define KEY_MINUS_PAD 74
#define KEY_PLUS_PAD 94
#define KEY_DEL_PAD 60
#define KEY_ENTER_PAD 67

#define KEY_NUMLOCK 90
#define KEY_SCRLOCK 91

#define KEY_LSHIFT 96
#define KEY_RSHIFT 97
#define KEY_CONTROL 99
#define KEY_LALT 100
#define KEY_RALT 101
#define KEY_CAPSLOCK 98

#define KEY_TAB 66
#define KEY_TILDE 0
#define KEY_MINUS 11
#define KEY_EQUALS 12
#define KEY_BACKSLASH 13
#define KEY_BACKSPACE 65

#define KEY_ENTER 68
#define KEY_COMMA 56
#define KEY_STOP 57
#define KEY_SLASH 58
#define KEY_SPACE 64

#define KEY_DEL 70
#define KEY_HELP 95

#define KEY_LAMIGA 102
#define KEY_RAMIGA 103

typedef struct {
  unsigned char ascii;
  unsigned char raw;
} key_info;

static key_info keylist[] =
{
  { 'A',  KEY_A },
  { 'B',  KEY_B },
  { 'C',  KEY_C },
  { 'D',  KEY_D },
  { 'E',  KEY_E },
  { 'F',  KEY_F },
  { 'G',  KEY_G },
  { 'H',  KEY_H },
  { 'I',  KEY_I },
  { 'J',  KEY_J },
  { 'K',  KEY_K },
  { 'L',  KEY_L },
  { 'M',  KEY_M },
  { 'N',  KEY_N },
  { 'O',  KEY_O },
  { 'P',  KEY_P },
  { 'Q',  KEY_Q },
  { 'R',  KEY_R },
  { 'S',  KEY_S },
  { 'T',  KEY_T },
  { 'U',  KEY_U },
  { 'V',  KEY_V },
  { 'W',  KEY_W },
  { 'X',  KEY_X },
  { 'Y',  KEY_Y },
  { 'Z',  KEY_Z },
  { '0',  KEY_0 },
  { '1',  KEY_1 },
  { '2',  KEY_2 },
  { '3',  KEY_3 },
  { '4',  KEY_4 },
  { '5',  KEY_5 },
  { '6',  KEY_6 },
  { '7',  KEY_7 },
  { '8',  KEY_8 },
  { '9',  KEY_9 },
  {   0,      0 } /* end of table */
};

static unsigned char keymap[128], key[128];

int check_idcmp(int get_char)
{
  struct IntuiMessage *imsg = NULL;
  ULONG imCode, imClass;
  int press = 0;

  if (get_char) {
    WaitPort(window->UserPort);
  }

  /* Check for IDCMP messages */
  if((imsg = (struct IntuiMessage *)GetMsg(window->UserPort))) {
    imClass = imsg->Class;
    imCode = imsg->Code;
    ReplyMsg((struct Message *)imsg);

    /* KeyPress and KeyReleases */
    if(imClass==IDCMP_RAWKEY) {
      if (imCode<128) { /* Key is being pressed */
        press = 1;
      }

      imCode &= 0x7F;

      if (press) {
        if (imCode == KEY_F1) {
          ui_fkeys |= 1 << 1;
        } else if (imCode == KEY_F2) {
          ui_fkeys |= 1 << 2;
        } else if (imCode == KEY_F3) {
          ui_fkeys |= 1 << 3;
        } else if (imCode == KEY_F4) {
          ui_fkeys |= 1 << 4;
        } else if (imCode == KEY_F5) {
          ui_fkeys |= 1 << 5;
        } else if (imCode == KEY_F6) {
          ui_fkeys |= 1 << 6;
        } else if (imCode == KEY_F7) {
          ui_fkeys |= 1 << 7;
        } else if (imCode == KEY_F8) {
          ui_fkeys |= 1 << 8;
        } else if (imCode == KEY_F9) {
          ui_fkeys |= 1 << 9;
        } else if (imCode == KEY_F10) {
          ui_fkeys |= 1 << 10;
        } else if (imCode == KEY_DEL) {
          ui_fkeys |= 1 << 11;
        } else if (imCode == KEY_HELP) {
          ui_fkeys |= 1 << 12;
        }
      }

      key[imCode] = press;

      if ((press) && (get_char)) {
        memset(key, 0, 128); /* Remove any keys to not affect the game */
        return keymap[imCode];
      }
    }
  }

  return 0;
}

/*** Code ***/

int uip_init(t_uipinfo * uipinfo)
{
  int i;

  uip_uipinfo = uipinfo;

  if (open_libraries() < 0) {
    close_libraries();
    return 1;
  }

  memset(key, 0, 128);

  /* Create keymap */
  memset(keymap, 0, 128);
  for (i=0; i<128; i++) {
    key_info key = keylist[i];
    if (key.ascii == 0) {
      break;
    }
    keymap[key.raw] = key.ascii;
  }

  return 0;
}

int uip_initjoysticks(void)
{
  if (LowLevelBase == NULL) {
    return 0;
  }

  /* Force CD32 pad (if required) */
  SetJoyPortAttrs(1, SJA_Type, SJA_TYPE_GAMECTLR, TAG_END);
  return 1;
}

int uip_vgamode(void)
{
  if (open_screen() < 0) {
    return 1;
  }

  atexit(close_screen);

  if (depth == 15) {
    uip_uipinfo->redshift = 10;
    uip_uipinfo->greenshift = 5;
    uip_uipinfo->blueshift = 0;
  } else {
    uip_uipinfo->redshift = 11;
    uip_uipinfo->greenshift = 6;
    uip_uipinfo->blueshift = 0;
  }

  uip_bank0 = (unsigned char *)baseaddress;
  uip_bank1 = uip_bank0 + (640 * 480 * 2);

  uip_uipinfo->screenmem0 = (uint8 *)uip_bank0;
  uip_uipinfo->screenmem1 = (uint8 *)uip_bank1;

  uip_uipinfo->linewidth = 2 * 640; /* 16 bit */
  uip_displaybank(0);           /* set current to 0th bank */
  uip_clearscreen();            /* clear bank */
  ui_setupscreen();             /* setup bank */
  uip_displaybank(-1);          /* toggle bank */
  uip_clearscreen();            /* clear bank */
  ui_setupscreen();             /* setup bank */

  return 0;
}

int uip_setcolourbits(int red, int green, int blue)
{
  return 0;
}

void uip_displaybank(int bank)
{
  int scroll;

  if (bank == -1)
    bank = uip_displaybanknum ^ 1;
  uip_writebank = bank ? uip_bank0 : uip_bank1;

  scroll = bank ? 480 : 0;
  screen->ViewPort.RasInfo->RyOffset = scroll;
  ScrollVPort(&screen->ViewPort);

  uip_uipinfo->screenmem_w = bank ? uip_uipinfo->screenmem0 :
    uip_uipinfo->screenmem1;
  uip_displaybanknum = bank;
}

void uip_clearscreen(void)
{
  int y;

  for (y = 0; y < 480; y++) {
    memset(uip_uipinfo->screenmem_w + uip_uipinfo->linewidth * y, 0, 640 * 2);
  }
}

void uip_clearmiddle(void)
{
  int i;

  for (i = 0; i < 240; i++) {
    memset(uip_uipinfo->screenmem_w +
           (uip_uipinfo->linewidth) * (120 + i) + 160 * 2, 0, 2 * 320);
  }
}

/* uip_singlebank sets write address to the current display screen */

void uip_singlebank(void)
{
  uip_uipinfo->screenmem_w = uip_displaybanknum ? uip_uipinfo->screenmem1 :
    uip_uipinfo->screenmem0;
}

/* uip_doublebank sets write address to hidden display screen */

void uip_doublebank(void)
{
  uip_uipinfo->screenmem_w = uip_displaybanknum ? uip_uipinfo->screenmem0 :
    uip_uipinfo->screenmem1;
}

void uip_textmode(void)
{
}

int uip_checkkeyboard(void)
{
  int pad;
  int joystick;

  check_idcmp(0);

  for (pad = 0; pad < 2; pad++) {
    if ((joystick = ui_bindings[pad].joystick) >= 0) {
      unsigned long portstate;

      /* Read port 1 (joyport) */
      portstate=ReadJoyPort(1);

      /* Was any buttons pressed? */
      if (portstate&JPF_JOY_LEFT) /* Left */
        mem68k_cont[pad].left = 1;
      else
        mem68k_cont[pad].left = 0;

      if (portstate&JPF_JOY_RIGHT) /* Right */
        mem68k_cont[pad].right = 1;
      else
        mem68k_cont[pad].right = 0;

      if (portstate&JPF_JOY_UP) /* Up */
        mem68k_cont[pad].up = 1;
      else
        mem68k_cont[pad].up = 0;

      if (portstate&JPF_JOY_DOWN) /* Down */
        mem68k_cont[pad].down = 1;
      else
        mem68k_cont[pad].down = 0;

      if (portstate&JPF_BUTTON_RED) /* A */
        mem68k_cont[pad].a = 1;
      else
        mem68k_cont[pad].a = 0;

      if (portstate&JPF_BUTTON_BLUE) /* B */
        mem68k_cont[pad].b = 1;
      else
        mem68k_cont[pad].b = 0;

      if (portstate&JPF_BUTTON_YELLOW) /* C */
        mem68k_cont[pad].c = 1;
      else
        mem68k_cont[pad].c = 0;

      if (portstate&JPF_BUTTON_PLAY) /* Start */
        mem68k_cont[pad].start = 1;
      else
        mem68k_cont[pad].start = 0;
/*
      JPF_BUTTON_FORWARD Charcoal - Forward
      JPF_BUTTON_REVERSE Charcoal - Reverse
      JPF_BUTTON_GREEN - Green Button
*/
    } else {
      switch (ui_bindings[pad].keyboard) {
      case 0: /* main keys */
        mem68k_cont[pad].a = key[KEY_A] ? 1 : 0;
        mem68k_cont[pad].b = (key[KEY_B] ||
                              key[KEY_S]) ? 1 : 0;
        mem68k_cont[pad].c = (key[KEY_C] ||
                              key[KEY_D]) ? 1 : 0;
        mem68k_cont[pad].left = key[KEY_LEFT] ? 1 : 0;
        mem68k_cont[pad].up = key[KEY_UP] ? 1 : 0;
        mem68k_cont[pad].right = key[KEY_RIGHT] ? 1 : 0;
        mem68k_cont[pad].down = key[KEY_DOWN] ? 1 : 0;
        mem68k_cont[pad].start = key[KEY_ENTER] ? 1 : 0;
        break;
      case 1: /* left side of keyboard */
        mem68k_cont[pad].a = key[KEY_Z] ? 1 : 0;
        mem68k_cont[pad].b = key[KEY_X] ? 1 : 0;
        mem68k_cont[pad].c = key[KEY_C] ? 1 : 0;
        mem68k_cont[pad].left = key[KEY_D] ? 1 : 0;
        mem68k_cont[pad].up = key[KEY_R] ? 1 : 0;
        mem68k_cont[pad].right = key[KEY_G] ? 1 : 0;
        mem68k_cont[pad].down = key[KEY_F] ? 1 : 0;
        mem68k_cont[pad].start = key[KEY_V] ? 1 : 0;
        break;
      case 2: /* right side of keyboard */
        mem68k_cont[pad].a = key[KEY_COMMA] ? 1 : 0;
        mem68k_cont[pad].b = key[KEY_STOP] ? 1 : 0;
        mem68k_cont[pad].c = key[KEY_SLASH] ? 1 : 0;
        mem68k_cont[pad].left = key[KEY_LEFT] ? 1 : 0;
        mem68k_cont[pad].up = key[KEY_UP] ? 1 : 0;
        mem68k_cont[pad].right = key[KEY_RIGHT] ? 1 : 0;
        mem68k_cont[pad].down = key[KEY_DOWN] ? 1 : 0;
        mem68k_cont[pad].start = key[KEY_ENTER] ? 1 : 0;
        break;
      }
    }
  }
  return (key[KEY_ESC] ? 1 : 0);
}

unsigned int uip_whichbank(void)
{
  /* returns 0 or 1 - the bank being VIEWED */
  return uip_displaybanknum;
}

void uip_vsync(void)
{
  WaitTOF(); /* VSync */
}

uint8 uip_getchar(void)
{
  char c = 0;

  while (c == 0) {
    c = check_idcmp(1);
  }

  return c;
}
